"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ymir_1 = require("@fimbul/ymir");
const chalk_1 = require("chalk");
const path = require("path");
const COLORS = {
    error: chalk_1.default.red,
    warning: chalk_1.default.yellow,
    suggestion: chalk_1.default.cyan,
};
const SYMBOLS = {
    error: '✖',
    warning: '⚠',
    suggestion: '💡',
};
class Formatter extends ymir_1.AbstractFormatter {
    constructor() {
        super(...arguments);
        this.fixed = 0;
        this.fixable = 0;
        this.maxSeverityWidth = 0;
        this.maxPositionWidth = 0;
        this.maxNameWidth = 0;
        this.files = new Map();
    }
    format(fileName, summary) {
        this.fixed += summary.fixes;
        if (summary.findings.length === 0)
            return;
        const mapped = [];
        for (const finding of summary.findings.slice().sort(ymir_1.Finding.compare)) {
            if (finding.fix !== undefined)
                ++this.fixable;
            if (finding.severity.length > this.maxSeverityWidth)
                this.maxSeverityWidth = finding.severity.length;
            if (finding.ruleName.length > this.maxNameWidth)
                this.maxNameWidth = finding.ruleName.length;
            let { character, line } = finding.start;
            if (line !== 0 || character === 0 || !summary.content.startsWith('\uFEFF'))
                character += 1; // avoid incrementing the character position on the first line if BOM is present, editors ignore BOM
            const position = `${line + 1}:${character}`;
            if (position.length > this.maxPositionWidth)
                this.maxPositionWidth = position.length;
            mapped.push({
                position,
                severity: finding.severity,
                ruleName: finding.ruleName,
                message: finding.message,
            });
        }
        this.files.set(fileName, mapped);
        return;
    }
    flush() {
        const counts = {
            error: 0,
            warning: 0,
            suggestion: 0,
        };
        const lines = [];
        for (const [fileName, findings] of this.files) {
            lines.push('', `${chalk_1.default.underline(path.normalize(fileName))}${chalk_1.default.hidden(':' + findings[0].position)}`);
            for (const finding of findings) {
                const positionColor = COLORS[finding.severity];
                ++counts[finding.severity];
                lines.push(positionColor(pad(finding.severity.toUpperCase(), this.maxSeverityWidth) + ' ' + pad(finding.position, this.maxPositionWidth)) + `  ${chalk_1.default.grey(pad(finding.ruleName, this.maxNameWidth))}  ${chalk_1.default.blueBright(finding.message)}`);
            }
        }
        if (this.fixed !== 0)
            lines.push('', chalk_1.default.green(`Automatically fixed ${addCount(this.fixed, 'finding')}.`));
        if (this.files.size !== 0) {
            const summaryLine = [];
            for (const severity of Object.keys(counts)) {
                const count = counts[severity];
                if (count !== 0)
                    summaryLine.push(COLORS[severity].bold(`${SYMBOLS[severity]} ${addCount(count, severity)}`));
            }
            lines.push('', summaryLine.join('  '));
            if (this.fixable !== 0)
                lines.push(chalk_1.default.grey(addCount(this.fixable, 'finding') + ' ' +
                    (this.fixable === 1 ? 'is' : 'are') + " potentially fixable with the '--fix' option."));
        }
        return lines.length === 0
            ? undefined
            : lines.slice(1).join('\n'); // remove first line, because it's always empty
    }
}
exports.Formatter = Formatter;
function addCount(count, word) {
    return `${count} ${word}${count === 1 ? '' : 's'}`;
}
function pad(str, width) {
    return str + ' '.repeat(width - str.length);
}
//# sourceMappingURL=stylish.js.map