"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
let Rule = class Rule extends ymir_1.TypedRule {
    apply() {
        this.context.getWrappedAst().children.forEach(this.checkNode, this);
    }
    checkNode({ node, children }) {
        if (tsutils_1.isTypeNodeKind(node.kind))
            return;
        if (tsutils_1.isExpression(node)) {
            if (!tsutils_1.isIdentifier(node) || tsutils_1.getUsageDomain(node) !== undefined)
                this.checkAssignment(node);
        }
        else if (tsutils_1.isPropertyDeclaration(node)) {
            this.checkClassProperty(node, node.parent);
        }
        else if (tsutils_1.isMethodDeclaration(node)) {
            const parent = node.parent;
            if (parent.kind === ts.SyntaxKind.ObjectLiteralExpression) {
                this.checkObjectMethodDeclaration(node, parent);
            }
            else {
                this.checkClassProperty(node, parent);
            }
        }
        return children.forEach(this.checkNode, this);
    }
    checkAssignment(node) {
        if (returnTypeMatches(this.checker.getContextualType(node), this.checker, isVoidType) &&
            returnTypeMatches(this.checker.getTypeAtLocation(node), this.checker, typeContainsThenable, node)) {
            let errorNode = node;
            if (tsutils_1.isArrowFunction(node)) {
                errorNode = tsutils_1.getModifier(node, ts.SyntaxKind.AsyncKeyword) || node.equalsGreaterThanToken;
            }
            else if (tsutils_1.isFunctionExpression(node)) {
                errorNode = tsutils_1.getModifier(node, ts.SyntaxKind.AsyncKeyword) ||
                    node.name ||
                    tsutils_1.getChildOfKind(node, ts.SyntaxKind.FunctionKeyword);
            }
            this.addFindingAtNode(errorNode, "A 'Promise'-returning function should not be assigned to a 'void'-returning function type.");
        }
    }
    checkClassProperty(node, clazz) {
        if (clazz.heritageClauses === undefined)
            return;
        const checker = this.checker;
        if (node.kind === ts.SyntaxKind.MethodDeclaration) {
            const signature = checker.getSignatureFromDeclaration(node);
            if (signature === undefined || !typeContainsThenable(signature.getReturnType(), checker, clazz))
                return;
        }
        else if (!returnTypeMatches(checker.getTypeAtLocation(node), checker, typeContainsThenable, clazz)) {
            return;
        }
        const property = tsutils_1.getSingleLateBoundPropertyNameOfPropertyName(node.name, this.checker);
        if (property === undefined)
            return;
        const { displayName, symbolName } = property;
        for (const heritageClause of clazz.heritageClauses)
            for (const base of heritageClause.types)
                if (returnTypeMatches(this.getTypeOfProperty(checker.getTypeAtLocation(base), symbolName, base), checker, isVoidType))
                    return this.addFindingAtNode(tsutils_1.getModifier(node, ts.SyntaxKind.AsyncKeyword) || node.name, `Overriding 'void'-returning method '${displayName}' of base type with a 'Promise'-returning method is unsafe.`);
    }
    getTypeOfProperty(classType, name, node) {
        const propertySymbol = tsutils_1.getPropertyOfType(classType, name);
        return propertySymbol && this.checker.getTypeOfSymbolAtLocation(propertySymbol, node);
    }
    checkObjectMethodDeclaration(node, parent) {
        const contextualType = this.checker.getContextualType(parent);
        if (contextualType === undefined)
            return;
        for (const { displayName, symbolName } of tsutils_1.getLateBoundPropertyNamesOfPropertyName(node.name, this.checker).names) {
            const property = tsutils_1.getPropertyOfType(contextualType, symbolName);
            const propertyType = property !== undefined
                ? this.checker.getTypeOfSymbolAtLocation(property, parent)
                : tsutils_1.isNumericPropertyName(symbolName) && contextualType.getNumberIndexType() || contextualType.getStringIndexType();
            if (!returnTypeMatches(propertyType, this.checker, isVoidType))
                continue;
            const signature = this.checker.getSignatureFromDeclaration(node);
            if (signature !== undefined && typeContainsThenable(signature.getReturnType(), this.checker, node))
                return this.addFindingAtNode(tsutils_1.getModifier(node, ts.SyntaxKind.AsyncKeyword) || node.name, `'Promise'-returning method '${displayName}' should not be assigned to a 'void'-returning function type.`);
        }
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
function returnTypeMatches(type, checker, predicate, param) {
    if (type === undefined)
        return false;
    const callSignatures = tsutils_1.removeOptionalityFromType(checker, type).getCallSignatures();
    return callSignatures.length !== 0 && callSignatures.every((signature) => predicate(signature.getReturnType(), checker, param));
}
function typeContainsThenable(type, checker, node) {
    return tsutils_1.unionTypeParts(type).some((t) => tsutils_1.isThenableType(checker, node, t));
}
function isVoidType(type) {
    return (type.flags & ts.TypeFlags.Void) !== 0;
}
//# sourceMappingURL=async-function-assignability.js.map