"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
const utils_1 = require("../utils");
let Rule = class Rule extends ymir_1.TypedRule {
    apply() {
        const re = /\bawait\b/g;
        let wrappedAst;
        for (let match = re.exec(this.sourceFile.text); match !== null; match = re.exec(this.sourceFile.text)) {
            const { node } = tsutils_1.getWrappedNodeAtPosition(wrappedAst || (wrappedAst = this.context.getWrappedAst()), match.index);
            if (tsutils_1.isAwaitExpression(node)) {
                if (node.expression.pos !== re.lastIndex ||
                    this.maybePromiseLike(this.checker.getTypeAtLocation(node.expression), node.expression))
                    continue;
                const fix = [ymir_1.Replacement.delete(match.index, node.expression.getStart(this.sourceFile))];
                if (utils_1.expressionNeedsParensWhenReplacingNode(node.expression, node))
                    fix.push(ymir_1.Replacement.append(match.index, '('), ymir_1.Replacement.append(node.expression.end, ')'));
                this.addFinding(match.index, node.end, "Unnecessary 'await' of a non-Promise value.", fix);
            }
            else if (node.kind === ts.SyntaxKind.AwaitKeyword && node.end === re.lastIndex) {
                const parent = node.parent;
                if (tsutils_1.isForOfStatement(parent) && !this.isAsyncIterable(parent.expression)) {
                    const start = node.pos - 'for'.length;
                    this.addFinding(start, parent.statement.pos, "Unnecessary 'for await' of a non-AsyncIterable value.", ymir_1.Replacement.delete(start + 'for'.length, re.lastIndex));
                }
            }
        }
    }
    maybePromiseLike(type, node) {
        type = this.checker.getApparentType(type);
        if (type.flags & (ts.TypeFlags.Any | ts.TypeFlags.Unknown))
            return true;
        for (const t of tsutils_1.unionTypeParts(type))
            if (this.isThenable(t, node))
                return true;
        return false;
    }
    /**
     * A type is thenable when it has a callable `then` property.
     * We don't care if the signatures are actually callable because the compiler already complains about that.
     */
    isThenable(type, node) {
        const then = type.getProperty('then');
        return then !== undefined && this.checker.getTypeOfSymbolAtLocation(then, node).getCallSignatures().length !== 0;
    }
    isAsyncIterable(node) {
        const type = this.checker.getApparentType(this.checker.getTypeAtLocation(node));
        if (type.flags & (ts.TypeFlags.Any | ts.TypeFlags.Unknown))
            return true;
        for (const t of tsutils_1.unionTypeParts(type))
            // there must either be a `AsyncIterable` or `Iterable<PromiseLike<any>>`
            if (this.hasSymbolAsyncIterator(t) || this.isIterableOfPromises(t, node))
                return true;
        return false;
    }
    /**
     * We consider a type as AsyncIterable when it has a property key [Symbol.asyncIterator].
     * The spec requires this property to be a function returning an object with a `next` method which returns a Promise of IteratorResult.
     * But that's none of our business as the compiler already does the heavy lifting.
     */
    hasSymbolAsyncIterator(type) {
        return type.getProperties().some((prop) => prop.escapedName === '__@asyncIterator');
    }
    isIterableOfPromises(type, node) {
        const symbol = type.getProperties().find((prop) => prop.escapedName === '__@iterator');
        if (symbol === undefined)
            return false;
        const t = this.checker.getTypeOfSymbolAtLocation(symbol, node);
        if (t.flags & ts.TypeFlags.Any)
            return true;
        for (const signature of t.getCallSignatures()) {
            const returnType = signature.getReturnType();
            if (returnType.flags & ts.TypeFlags.Any)
                return true;
            const next = returnType.getProperty('next');
            if (next === undefined)
                continue;
            const nextType = this.checker.getTypeOfSymbolAtLocation(next, node);
            if (nextType.flags & ts.TypeFlags.Any)
                return true;
            for (const nextSignature of nextType.getCallSignatures()) {
                const nextReturnType = nextSignature.getReturnType();
                if (nextReturnType.flags & ts.TypeFlags.Any)
                    return true;
                const value = nextReturnType.getProperty('value');
                if (value !== undefined && this.maybePromiseLike(this.checker.getTypeOfSymbolAtLocation(value, node), node))
                    return true;
            }
        }
        return false;
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
//# sourceMappingURL=await-only-promise.js.map