"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
const emptyPropertyInfo = {
    known: false,
    names: [],
    assignedNames: [],
};
let Rule = class Rule extends ymir_1.TypedRule {
    apply() {
        const checkedObjects = new Set();
        for (const node of this.context.getFlatAst()) {
            if (node.kind === ts.SyntaxKind.SpreadAssignment &&
                !checkedObjects.has(node.parent.pos) &&
                !tsutils_1.isReassignmentTarget(node.parent)) {
                checkedObjects.add(node.parent.pos);
                this.checkObject(node.parent);
            }
        }
    }
    checkObject({ properties }) {
        /** key: propertyName, value: isAccessor */
        const propertiesSeen = new Map();
        for (let i = properties.length - 1; i >= 0; --i) {
            const property = properties[i];
            const info = this.getPropertyInfo(property);
            const isAccessor = property.kind === ts.SyntaxKind.GetAccessor || property.kind === ts.SyntaxKind.SetAccessor;
            if (info.known && info.names.every((name) => isAccessor ? propertiesSeen.get(name) === false : propertiesSeen.has(name))) {
                if (property.kind === ts.SyntaxKind.SpreadAssignment) {
                    this.addFindingAtNode(property, 'All properties of this object are overridden later.');
                }
                else {
                    this.addFindingAtNode(property.name, `Property '${property.name.getText(this.sourceFile)}' is overridden later.`);
                    if (isAccessor)
                        continue; // avoid overriding the isAccessor state
                }
            }
            for (const name of info.assignedNames)
                propertiesSeen.set(name, isAccessor);
        }
    }
    getPropertyInfo(property) {
        switch (property.kind) {
            case ts.SyntaxKind.SpreadAssignment:
                return this.getPropertyInfoFromSpread(property.expression);
            case ts.SyntaxKind.ShorthandPropertyAssignment:
                return {
                    known: true,
                    names: [property.name.escapedText],
                    assignedNames: [property.name.escapedText],
                };
            default: {
                const lateBound = tsutils_1.getLateBoundPropertyNamesOfPropertyName(property.name, this.checker);
                if (!lateBound.known)
                    return emptyPropertyInfo;
                const names = lateBound.names.map((p) => p.symbolName);
                return {
                    names,
                    known: true,
                    assignedNames: names.length !== 1 ? [] : names,
                };
            }
        }
    }
    getPropertyInfoFromSpread(node) {
        return getPropertyInfoFromType(this.checker.getTypeAtLocation(node));
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles,
    ymir_1.requiresCompilerOption('strictNullChecks')
], Rule);
exports.Rule = Rule;
function getPropertyInfoFromType(type) {
    if (tsutils_1.isUnionType(type))
        return type.types.map(getPropertyInfoFromType).reduce(unionPropertyInfo);
    if (tsutils_1.isIntersectionType(type))
        return type.types.map(getPropertyInfoFromType).reduce(intersectPropertyInfo);
    if (type.flags & ts.TypeFlags.Instantiable) {
        const constraint = type.getConstraint();
        if (constraint === undefined)
            return emptyPropertyInfo;
        return Object.assign({}, getPropertyInfoFromType(constraint), { known: false });
    }
    if (!tsutils_1.isObjectType(type))
        return emptyPropertyInfo;
    const result = {
        known: type.getStringIndexType() === undefined && type.getNumberIndexType() === undefined,
        names: [],
        assignedNames: [],
    };
    for (const prop of type.getProperties()) {
        if (!isSpreadableProperty(prop))
            continue;
        if ((prop.flags & ts.SymbolFlags.Optional) === 0)
            result.assignedNames.push(prop.escapedName);
        result.names.push(prop.escapedName);
    }
    return result;
}
function isSpreadableProperty(prop) {
    if (prop.flags & (ts.SymbolFlags.Method | ts.SymbolFlags.Accessor))
        for (const declaration of prop.declarations)
            if (tsutils_1.isClassLikeDeclaration(declaration.parent))
                return false;
    return true;
}
function unionPropertyInfo(a, b) {
    return {
        known: a.known && b.known,
        names: [...a.names, ...b.names],
        assignedNames: a.assignedNames.filter((name) => b.assignedNames.includes(name)),
    };
}
function intersectPropertyInfo(a, b) {
    return {
        known: a.known && b.known,
        names: [...a.names, ...b.names],
        assignedNames: [...a.assignedNames, ...b.assignedNames],
    };
}
//# sourceMappingURL=no-duplicate-spread-property.js.map