"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const utils_1 = require("../utils");
const tsutils_1 = require("tsutils");
let Rule = class Rule extends ymir_1.TypedRule {
    apply() {
        for (const node of this.context.getFlatAst())
            if (node.kind === ts.SyntaxKind.ElementAccessExpression)
                this.checkElementAccess(node);
    }
    checkElementAccess(node) {
        const { names } = tsutils_1.getLateBoundPropertyNames(node.argumentExpression, this.checker);
        if (names.length === 0)
            return;
        const type = this.checker.getApparentType(this.checker.getTypeAtLocation(node.expression));
        for (const { symbol, name } of utils_1.propertiesOfType(type, names))
            this.checkSymbol(symbol, name, node, node.expression, type);
    }
    checkSymbol(symbol, name, errorNode, lhs, lhsType) {
        const flags = getModifierFlagsOfSymbol(symbol);
        if (hasConflictingAccessModifiers(flags, symbol))
            return this.addFindingAtNode(errorNode, `Property '${name}' has conflicting declarations and is inaccessible in type '${this.checker.typeToString(lhsType)}'.`);
        if (lhs !== undefined && lhs.kind === ts.SyntaxKind.ThisKeyword &&
            flags & ts.ModifierFlags.Abstract && hasNonMethodDeclaration(symbol)) {
            const enclosingClass = getEnclosingClassOfAbstractPropertyAccess(errorNode.parent);
            if (enclosingClass !== undefined)
                return this.addFindingAtNode(errorNode, `Abstract property '${name}' in class '${this.printClass(enclosingClass)}' cannot be accessed during class initialization.`);
        }
        if (lhs !== undefined && lhs.kind === ts.SyntaxKind.SuperKeyword) {
            if (hasNonMethodDeclaration(symbol))
                return this.addFindingAtNode(errorNode, "Only public and protected methods of the base class are accessible via the 'super' keyword.");
            if (flags & ts.ModifierFlags.Abstract &&
                symbol.declarations.every((d) => tsutils_1.hasModifier(d.modifiers, ts.SyntaxKind.AbstractKeyword)))
                return this.addFindingAtNode(errorNode, `Abstract method '${name}' in class '${this.printClass(symbol.declarations[0].parent)}' cannot be accessed via the 'super' keyword.`);
        }
        if ((flags & ts.ModifierFlags.NonPublicAccessibilityModifier) === 0)
            return;
        if (flags & ts.ModifierFlags.Private) {
            const declaringClass = symbol.declarations[0].parent;
            if (errorNode.pos < declaringClass.pos || errorNode.end > declaringClass.end)
                this.failVisibility(errorNode, name, this.printClass(declaringClass), true);
        }
        else {
            const declaringClasses = symbol.declarations.map((d) => d.parent);
            let enclosingClass = this.findEnclosingClass(errorNode.parent.parent, declaringClasses);
            if (enclosingClass === undefined) {
                if ((flags & ts.ModifierFlags.Static) === 0)
                    enclosingClass = this.getEnclosingClassFromThisParameter(errorNode.parent.parent, declaringClasses);
                if (enclosingClass === undefined)
                    return this.failVisibility(errorNode, name, this.checker.typeToString(lhsType), false);
            }
            if ((flags & ts.ModifierFlags.Static) === 0 && !hasBase(lhsType, enclosingClass, isIdentical))
                return this.addFindingAtNode(errorNode, `Property '${name}' is protected and only accessible through an instance of class '${this.checker.typeToString(enclosingClass)}'.`);
        }
    }
    getEnclosingClassFromThisParameter(node, baseClasses) {
        const thisParameter = getThisParameterFromContext(node);
        if (thisParameter === undefined || thisParameter.type === undefined)
            return;
        let thisType = this.checker.getTypeFromTypeNode(thisParameter.type);
        if (tsutils_1.isTypeParameter(thisType)) {
            const constraint = thisType.getConstraint();
            if (constraint === undefined)
                return;
            thisType = constraint;
        }
        if (tsutils_1.isTypeReference(thisType))
            thisType = thisType.target;
        return baseClasses.every((baseClass) => hasBase(thisType, baseClass, typeContainsDeclaration)) ? thisType : undefined;
    }
    failVisibility(node, property, typeString, isPrivate) {
        this.addFindingAtNode(node, `Property '${property}' is ${isPrivate ? 'private' : 'protected'} and only accessible within class '${typeString}'${isPrivate ? '' : ' and its subclasses'}.`);
    }
    findEnclosingClass(node, baseClasses) {
        while (true) {
            switch (node.kind) {
                case ts.SyntaxKind.ClassDeclaration:
                case ts.SyntaxKind.ClassExpression: {
                    const declaredType = tsutils_1.getConstructorTypeOfClassLikeDeclaration(node, this.checker);
                    if (baseClasses.every((baseClass) => hasBase(declaredType, baseClass, typeContainsDeclaration)))
                        return declaredType;
                    break;
                }
                case ts.SyntaxKind.SourceFile:
                    return;
            }
            node = node.parent;
        }
    }
    printClass(declaration) {
        return this.checker.typeToString(tsutils_1.getConstructorTypeOfClassLikeDeclaration(declaration, this.checker));
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
function hasNonMethodDeclaration(symbol) {
    return !symbol.declarations.every(tsutils_1.isMethodDeclaration);
}
function getEnclosingClassOfAbstractPropertyAccess(node) {
    while (true) {
        if (tsutils_1.isFunctionScopeBoundary(node)) {
            switch (node.kind) {
                case ts.SyntaxKind.ClassDeclaration:
                case ts.SyntaxKind.ClassExpression:
                    return node;
                case ts.SyntaxKind.Constructor:
                    return node.parent;
                default:
                    return;
            }
        }
        node = node.parent;
    }
}
function hasConflictingAccessModifiers(flags, symbol) {
    return flags & ts.ModifierFlags.Private
        ? symbol.declarations.length !== 1
        : (flags & ts.ModifierFlags.Protected) !== 0 &&
            symbol.declarations.length !== 1 &&
            !symbol.declarations.every((d) => tsutils_1.hasModifier(d.modifiers, ts.SyntaxKind.ProtectedKeyword));
}
function hasBase(type, needle, check) {
    return (function recur(t) {
        if (tsutils_1.isTypeReference(t)) {
            t = t.target;
            if (check(t, needle))
                return true;
        }
        const baseTypes = t.getBaseTypes();
        if (baseTypes !== undefined && baseTypes.some(recur))
            return true;
        return tsutils_1.isIntersectionType(t) && t.types.some(recur);
    })(type);
}
function isIdentical(a, b) {
    return a === b;
}
function typeContainsDeclaration(type, declaration) {
    return type.symbol !== undefined && type.symbol.declarations !== undefined && type.symbol.declarations.includes(declaration);
}
function getThisParameterFromContext(node) {
    while (true) {
        switch (node.kind) {
            case ts.SyntaxKind.FunctionDeclaration:
            case ts.SyntaxKind.FunctionExpression:
            case ts.SyntaxKind.MethodDeclaration: {
                const { parameters } = node;
                return parameters.length !== 0 && tsutils_1.isThisParameter(parameters[0]) ? parameters[0] : undefined;
            }
            case ts.SyntaxKind.ClassDeclaration:
            case ts.SyntaxKind.ClassExpression:
            case ts.SyntaxKind.SetAccessor:
            case ts.SyntaxKind.GetAccessor:
            case ts.SyntaxKind.EnumDeclaration:
            case ts.SyntaxKind.SourceFile:
                return;
            case ts.SyntaxKind.Decorator:
                // skip the declaration the decorator belongs to, because decorators are always applied in the outer context
                node = node.parent.parent;
                break;
            default:
                node = node.parent;
        }
    }
}
function getModifierFlagsOfSymbol(symbol) {
    return symbol.declarations === undefined
        ? ts.ModifierFlags.None
        : symbol.declarations.reduce((flags, decl) => flags | ts.getCombinedModifierFlags(decl), ts.ModifierFlags.None);
}
//# sourceMappingURL=no-restricted-property-access.js.map