"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
const emptyObjectFallback = /^3\.[234]$/.test(ts.versionMajorMinor);
const emptyObjectType = {
    display: '{}',
    predicate(type) {
        return tsutils_1.isTypeLiteralNode(type) && type.members.length === 0;
    },
};
const unknownType = {
    display: 'unknown',
    predicate(type) {
        return type.kind === ts.SyntaxKind.UnknownKeyword;
    },
};
const anyType = {
    display: 'any',
    predicate(type) {
        return type.kind === ts.SyntaxKind.AnyKeyword;
    },
};
let Rule = class Rule extends ymir_1.TypedRule {
    constructor() {
        super(...arguments);
        this.checkType = !/\.tsx?$/.test(this.sourceFile.fileName) ? anyType : emptyObjectFallback ? emptyObjectType : unknownType;
    }
    apply() {
        for (const node of this.context.getFlatAst()) {
            if (node.kind === ts.SyntaxKind.CallExpression) {
                if (node.typeArguments === undefined)
                    this.checkCallExpression(node);
            }
            else if (node.kind === ts.SyntaxKind.NewExpression) {
                if (node.typeArguments === undefined)
                    this.checkNewExpression(node);
            }
            else if ((node.kind === ts.SyntaxKind.JsxOpeningElement ||
                node.kind === ts.SyntaxKind.JsxSelfClosingElement ||
                node.kind === ts.SyntaxKind.TaggedTemplateExpression) &&
                node.typeArguments === undefined) {
                this.checkCallExpression(node);
            }
        }
    }
    /**
     * This function is necessary because higher order function type inference creates Signatures whose declaration has no type parameters.
     * @see https://github.com/Microsoft/TypeScript/issues/30296
     *
     * To work around this, we look for a single call signature on the called expression and use its type parameters instead.
     * As a bonus this also works for unified signatures from union types.
     */
    getTypeParametersOfCallSignature(node) {
        let expr;
        switch (node.kind) {
            case ts.SyntaxKind.CallExpression:
            case ts.SyntaxKind.NewExpression:
                expr = node.expression;
                break;
            case ts.SyntaxKind.TaggedTemplateExpression:
                expr = node.tag;
                break;
            default:
                expr = node.tagName;
        }
        const signatures = this.checker.getSignaturesOfType(this.checker.getTypeAtLocation(expr), node.kind === ts.SyntaxKind.NewExpression ? ts.SignatureKind.Construct : ts.SignatureKind.Call);
        // abort if not all signatures have type parameters:
        //   higher order function type inference only works for a single call signature
        //   call signature unification puts type parameters on every resulting signature
        if (signatures.length === 0 || signatures.some((s) => s.typeParameters === undefined))
            return [];
        return signatures[0].typeParameters.map(this.mapTypeParameter, this);
    }
    mapTypeParameter(type) {
        // fall back to NodeBuilder for renamed TypeParameters, they have no declaration and therefore we cannot directly access the default
        return type.symbol.declarations === undefined
            ? mapTypeParameterDeclaration(this.checker.typeParameterToDeclaration(type, undefined, ts.NodeBuilderFlags.IgnoreErrors))
            : {
                name: type.symbol.name,
                hasDefault: type.symbol.declarations[0].default !== undefined,
            };
    }
    checkCallExpression(node) {
        const signature = this.checker.getResolvedSignature(node);
        if (signature.declaration !== undefined) {
            let typeParameters = ts.getEffectiveTypeParameterDeclarations(signature.declaration).map(mapTypeParameterDeclaration);
            if (typeParameters.length === 0)
                typeParameters = this.getTypeParametersOfCallSignature(node);
            if (typeParameters.length !== 0)
                return this.checkInferredTypeParameters(signature, typeParameters, node);
        }
    }
    checkNewExpression(node) {
        const signature = this.checker.getResolvedSignature(node);
        if (signature.declaration !== undefined) {
            // There is an explicitly declared construct signature
            let typeParameters = ts.getEffectiveTypeParameterDeclarations(signature.declaration).map(mapTypeParameterDeclaration);
            if (typeParameters.length === 0)
                typeParameters = this.getTypeParametersOfCallSignature(node);
            if (typeParameters.length !== 0) // only check the signature if it declares type parameters
                return this.checkInferredTypeParameters(signature, typeParameters, node);
            if (signature.declaration.kind !== ts.SyntaxKind.Constructor)
                return; // don't look for type parameters on non-class parents
        }
        // Otherwise look up the TypeParameters of the ClassDeclaration
        const { symbol } = this.checker.getTypeAtLocation(node.expression);
        if (symbol === undefined || symbol.declarations === undefined)
            return;
        // collect all TypeParameters and their defaults from all merged declarations
        const mergedTypeParameters = [];
        for (const declaration of symbol.declarations) {
            const typeParameters = ts.getEffectiveTypeParameterDeclarations(declaration);
            for (let i = 0; i < typeParameters.length; ++i) {
                if (mergedTypeParameters.length === i)
                    mergedTypeParameters.push({ name: typeParameters[i].name.text, hasDefault: false });
                if (typeParameters[i].default !== undefined)
                    mergedTypeParameters[i].hasDefault = true;
            }
        }
        if (mergedTypeParameters.length !== 0)
            return this.checkInferredTypeParameters(signature, mergedTypeParameters, node);
    }
    checkInferredTypeParameters(signature, typeParameters, node) {
        if (typeParameters.every((t) => t.hasDefault))
            return; // nothing to do here if every type parameter as a default
        const typeArguments = this.checker.signatureToSignatureDeclaration(signature, ts.SyntaxKind.CallExpression, undefined, ts.NodeBuilderFlags.WriteTypeArgumentsOfSignature | ts.NodeBuilderFlags.IgnoreErrors).typeArguments;
        for (let i = 0; i < typeParameters.length; ++i)
            if (!typeParameters[i].hasDefault)
                this.checkTypeArgument(typeArguments[i], typeParameters[i].name, node);
    }
    checkTypeArgument(typeArgument, name, errorNode) {
        if (this.checkType.predicate(typeArgument))
            this.addFindingAtNode(errorNode, `TypeParameter '${name}' is inferred as '${this.checkType.display}'. Consider adding type arguments to the call.`);
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
function mapTypeParameterDeclaration(node) {
    return { name: node.name.text, hasDefault: node.default !== undefined };
}
//# sourceMappingURL=no-uninferred-type-parameter.js.map