"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
let Rule = class Rule extends ymir_1.AbstractRule {
    apply() {
        for (const node of this.context.getFlatAst()) {
            switch (node.kind) {
                case ts.SyntaxKind.Block:
                case ts.SyntaxKind.CaseClause:
                case ts.SyntaxKind.DefaultClause:
                    this.checkBlock(node);
                    break;
                case ts.SyntaxKind.IfStatement:
                    this.checkIfStatement(node);
                    break;
                case ts.SyntaxKind.ForStatement:
                case ts.SyntaxKind.WhileStatement:
                    this.checkConstantIterationCondition(node);
            }
        }
    }
    checkBlock(node) {
        let i = node.statements.findIndex(nextStatementIsUnreachable);
        if (i === -1 || i === node.statements.length - 1)
            return;
        for (i += 1; i < node.statements.length; ++i)
            if (isExecutableStatement(node.statements[i]))
                return this.report(node.statements[i]);
    }
    checkIfStatement(node) {
        switch (getConstantCondition(node.expression)) {
            case false:
                return this.report(node.thenStatement);
            case true:
                if (node.elseStatement !== undefined)
                    this.report(node.elseStatement);
        }
    }
    checkConstantIterationCondition(node) {
        if (getConstantIterationCondition(node) === false)
            this.report(node.statement);
    }
    report(node) {
        while (true) {
            if (tsutils_1.isLabeledStatement(node)) {
                node = node.statement;
            }
            else if (tsutils_1.isBlock(node)) {
                const next = node.statements.find(isExecutableStatement);
                if (next === undefined)
                    return;
                node = next;
            }
            else {
                break;
            }
        }
        this.addFindingAtNode(node.getFirstToken(this.sourceFile), 'Unreachable code detected.');
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
function nextStatementIsUnreachable(statement) {
    if (tsutils_1.endsControlFlow(statement))
        return true;
    const labels = [];
    while (tsutils_1.isLabeledStatement(statement)) {
        labels.push(statement.label.text);
        statement = statement.statement;
    }
    switch (statement.kind) {
        case ts.SyntaxKind.ForStatement:
        case ts.SyntaxKind.WhileStatement:
        case ts.SyntaxKind.DoStatement:
            // statements after loops, that go on forever without breaking, are never executed
            return getConstantIterationCondition(statement) === true &&
                !tsutils_1.getControlFlowEnd(statement.statement).statements.some((jump) => jump.kind === ts.SyntaxKind.BreakStatement && (jump.label === undefined || labels.includes(jump.label.text)));
    }
    return false;
}
function isExecutableStatement(node) {
    switch (node.kind) {
        case ts.SyntaxKind.FunctionDeclaration:
        case ts.SyntaxKind.InterfaceDeclaration:
        case ts.SyntaxKind.TypeAliasDeclaration:
        case ts.SyntaxKind.EmptyStatement:
            return false;
        case ts.SyntaxKind.EnumDeclaration:
            return !tsutils_1.hasModifier(node.modifiers, ts.SyntaxKind.ConstKeyword);
        case ts.SyntaxKind.VariableStatement:
            return tsutils_1.isBlockScopedVariableDeclarationList(node.declarationList) ||
                node.declarationList.declarations.some((d) => d.initializer !== undefined);
        case ts.SyntaxKind.Block:
            return node.statements.some(isExecutableStatement);
        case ts.SyntaxKind.LabeledStatement:
            return isExecutableStatement(node.statement);
        default:
            return true;
    }
}
function getConstantIterationCondition(statement) {
    return statement.kind === ts.SyntaxKind.ForStatement
        ? statement.condition === undefined || getConstantCondition(statement.condition)
        : getConstantCondition(statement.expression);
}
function getConstantCondition(node) {
    switch (node.kind) {
        case ts.SyntaxKind.TrueKeyword:
            return true;
        case ts.SyntaxKind.FalseKeyword:
            return false;
        default:
            return;
    }
}
//# sourceMappingURL=no-unreachable-code.js.map