"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ymir_1 = require("@fimbul/ymir");
const tsutils_1 = require("tsutils");
const ts = require("typescript");
const utils_1 = require("../utils");
const functionLikeSymbol = ts.SymbolFlags.Function | ts.SymbolFlags.Method;
const signatureFormatFlags = ts.TypeFormatFlags.UseFullyQualifiedType | ts.TypeFormatFlags.UseAliasDefinedOutsideCurrentScope;
class Rule extends ymir_1.TypedRule {
    apply() {
        for (const node of this.context.getFlatAst()) {
            // TODO maybe check Type["property"]
            if (tsutils_1.isIdentifier(node)) {
                if (shouldCheckIdentifier(node))
                    this.checkSymbol(this.checker.getSymbolAtLocation(node), node, node.text);
            }
            else if (tsutils_1.isPropertyAccessExpression(node)) {
                this.checkSymbol(this.checker.getSymbolAtLocation(node), node, node.name.text);
            }
            else if (tsutils_1.isElementAccessExpression(node)) {
                this.checkElementAccess(node);
            }
            else if (tsutils_1.isPropertyAssignment(node)) {
                if (tsutils_1.isReassignmentTarget(node.parent))
                    this.checkObjectDestructuring(node);
            }
            else if (tsutils_1.isShorthandPropertyAssignment(node)) {
                this.checkSymbol(this.checker.getShorthandAssignmentValueSymbol(node), node, node.name.text);
                if (tsutils_1.isReassignmentTarget(node.parent))
                    this.checkObjectDestructuring(node);
            }
            else if (tsutils_1.isCallLikeExpression(node)) {
                this.checkSignature(node);
            }
            else if (tsutils_1.isObjectBindingPattern(node)) {
                this.checkObjectBindingPattern(node);
            }
            else if (node.kind === ts.SyntaxKind.QualifiedName && shouldCheckQualifiedName(node)) {
                this.checkSymbol(this.checker.getSymbolAtLocation(node), node, node.right.text);
            }
        }
    }
    checkObjectDestructuring(node) {
        // for compatibility with typescript@<3.5.0
        if (this.checker.getTypeOfAssignmentPattern) {
            const type = this.checker.getTypeOfAssignmentPattern(node.parent);
            for (const { symbolName, displayName } of tsutils_1.getLateBoundPropertyNamesOfPropertyName(node.name, this.checker).names) {
                const symbol = tsutils_1.getPropertyOfType(type, symbolName);
                if (symbol !== undefined)
                    this.checkStability(symbol, node.name, displayName, describeWithName);
            }
        }
        else if (node.name.kind === ts.SyntaxKind.Identifier) {
            const symbol = this.checker.getPropertySymbolOfDestructuringAssignment(node.name);
            if (symbol !== undefined)
                return this.checkStability(symbol, node, node.name.text, describeWithName);
        }
    }
    checkSignature(node) {
        return this.checkStability(this.checker.getResolvedSignature(node), node, undefined, signatureToString);
    }
    checkObjectBindingPattern(node) {
        const type = this.checker.getTypeAtLocation(node);
        for (const element of node.elements) {
            if (element.dotDotDotToken !== undefined)
                continue;
            if (element.propertyName === undefined) {
                const name = element.name.text;
                const symbol = type.getProperty(name);
                if (symbol !== undefined)
                    this.checkStability(symbol, element.name, name, describeWithName);
            }
            else {
                const propName = tsutils_1.getPropertyName(element.propertyName);
                if (propName !== undefined) {
                    const symbol = type.getProperty(propName);
                    if (symbol !== undefined)
                        this.checkStability(symbol, element.propertyName, propName, describeWithName);
                }
                else {
                    for (const { symbol, name } of utils_1.propertiesOfType(type, tsutils_1.getLateBoundPropertyNames(element.propertyName.expression, this.checker).names))
                        this.checkStability(symbol, element.propertyName, name, describeWithName);
                }
            }
        }
    }
    checkElementAccess(node) {
        for (const { symbol, name } of utils_1.elementAccessSymbols(node, this.checker))
            this.checkSymbol(symbol, node, name);
    }
    checkSymbol(symbol, node, name) {
        if (symbol === undefined)
            return;
        if (symbol.flags & ts.SymbolFlags.Alias)
            symbol = this.checker.getAliasedSymbol(symbol);
        if ((symbol.flags & functionLikeSymbol) && isPartOfCall(node))
            return;
        return this.checkStability(symbol, node, name, describeWithName);
    }
    checkStability(s, node, hint, descr) {
        for (const tag of s.getJsDocTags())
            if (tag.name === 'deprecated' || tag.name === 'experimental')
                this.addFindingAtNode(node, `${descr(s, this.checker, hint, node)} is ${tag.name}${tag.text ? ': ' + tag.text : '.'}`);
    }
}
exports.Rule = Rule;
function describeWithName(symbol, _c, name) {
    return `${describeSymbol(symbol)} '${name}'`;
}
function describeSymbol(symbol) {
    if (symbol.flags & ts.SymbolFlags.Variable)
        return 'Variable';
    if (symbol.flags & ts.SymbolFlags.PropertyOrAccessor)
        return 'Property';
    if (symbol.flags & ts.SymbolFlags.Class)
        return 'Class';
    if (symbol.flags & ts.SymbolFlags.Enum)
        return 'Enum';
    if (symbol.flags & ts.SymbolFlags.EnumMember)
        return 'EnumMember';
    if (symbol.flags & ts.SymbolFlags.Function)
        return 'Function';
    if (symbol.flags & ts.SymbolFlags.Method)
        return 'Method';
    if (symbol.flags & ts.SymbolFlags.Interface)
        return 'Interface';
    if (symbol.flags & ts.SymbolFlags.NamespaceModule)
        return 'Namespace';
    if (symbol.flags & ts.SymbolFlags.TypeAlias)
        return 'TypeAlias';
    return '(unknown)';
}
function signatureToString(signature, checker, _, node) {
    let construct = false;
    switch (signature.declaration && signature.declaration.kind) {
        case ts.SyntaxKind.Constructor:
        case ts.SyntaxKind.ConstructSignature:
        case ts.SyntaxKind.ConstructorType:
            construct = true;
    }
    let name = '';
    const expr = getExpressionOfCallLike(node);
    if (tsutils_1.isIdentifier(expr)) {
        name = expr.text;
    }
    else if (tsutils_1.isPropertyAccessExpression(expr)) {
        name = expr.name.text;
    }
    else if (expr.kind === ts.SyntaxKind.SuperKeyword) {
        name = 'super';
    }
    return `${construct ? 'Costruct' : 'Call'}Signature '${construct && expr.kind !== ts.SyntaxKind.SuperKeyword ? 'new ' : ''}${name}${checker.signatureToString(signature, undefined, signatureFormatFlags)}'`;
}
function getExpressionOfCallLike(node) {
    switch (node.kind) {
        case ts.SyntaxKind.CallExpression:
        case ts.SyntaxKind.NewExpression:
        case ts.SyntaxKind.Decorator:
            return node.expression;
        case ts.SyntaxKind.TaggedTemplateExpression:
            return node.tag;
        case ts.SyntaxKind.JsxOpeningElement:
        case ts.SyntaxKind.JsxSelfClosingElement:
            return node.tagName;
    }
}
function isPartOfCall(node) {
    while (true) {
        const parent = node.parent;
        switch (parent.kind) {
            case ts.SyntaxKind.TaggedTemplateExpression:
            case ts.SyntaxKind.Decorator:
                return true;
            case ts.SyntaxKind.CallExpression:
                // note: NewExpression will never get here, because if the class is deprecated, we show an error all the time
                return parent.expression === node;
            case ts.SyntaxKind.JsxOpeningElement:
            case ts.SyntaxKind.JsxSelfClosingElement:
                return parent.tagName === node;
            case ts.SyntaxKind.ParenthesizedExpression:
                node = parent;
                break;
            default:
                return false;
        }
    }
}
function shouldCheckIdentifier(node) {
    switch (node.parent.kind) {
        case ts.SyntaxKind.ImportEqualsDeclaration:
        case ts.SyntaxKind.ExportAssignment:
        case ts.SyntaxKind.ExportSpecifier:
        case ts.SyntaxKind.JsxClosingElement:
            return false;
        case ts.SyntaxKind.ShorthandPropertyAssignment: // checked separately
            return node.parent.name !== node;
        default:
            return tsutils_1.getUsageDomain(node) !== undefined;
    }
}
function shouldCheckQualifiedName(node) {
    // if parent is a QualifiedName, it is the my.ns part of my.ns.Something -> we definitely want to check that
    // if the parent is an ImportEqualsDeclaration -> we don't want to check the rightmost identifier, because importing is not that bad
    // everything else is a TypeReference -> we want to check that
    return node.parent.kind !== ts.SyntaxKind.ImportEqualsDeclaration;
}
//# sourceMappingURL=no-unstable-api-use.js.map