"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
const utils_1 = require("../utils");
const FAIL_MESSAGE = 'This expression is unused. Did you mean to assign a value or call a function?';
let Rule = class Rule extends ymir_1.ConfigurableRule {
    parseOptions(input) {
        return Object.assign({ allowNew: false, allowShortCircuit: false, allowTaggedTemplate: false, allowTernary: false }, input);
    }
    apply() {
        for (const node of this.context.getFlatAst()) {
            if (tsutils_1.isBinaryExpression(node)) {
                if (node.operatorToken.kind === ts.SyntaxKind.CommaToken && !isIndirectEval(node))
                    this.checkNode(node.left);
            }
            else if (tsutils_1.isExpressionStatement(node)) {
                // allow `void asyncFn()` to mute 'await-async-result'
                if (!isDirective(node) && node.expression.kind !== ts.SyntaxKind.VoidExpression)
                    this.checkNode(node.expression, node);
            }
            else if (tsutils_1.isForStatement(node)) {
                if (node.initializer !== undefined && node.initializer.kind !== ts.SyntaxKind.VariableDeclarationList)
                    this.checkNode(node.initializer);
                if (node.incrementor !== undefined)
                    this.checkNode(node.incrementor);
            }
            else if (tsutils_1.isVoidExpression(node) && !isAllowedVoidExpression(node.expression)) {
                this.checkNode(node.expression);
            }
        }
    }
    checkNode(expr, errorNode) {
        if (!this.isUsed(expr))
            this.addFindingAtNode(errorNode === undefined ? expr : errorNode, FAIL_MESSAGE);
    }
    isUsed(node) {
        switch (node.kind) {
            case ts.SyntaxKind.CallExpression:
            case ts.SyntaxKind.YieldExpression:
            case ts.SyntaxKind.AwaitExpression:
            case ts.SyntaxKind.DeleteExpression:
            case ts.SyntaxKind.PostfixUnaryExpression: // i++, i--
                return true;
            case ts.SyntaxKind.PrefixUnaryExpression:
                return node.operator === ts.SyntaxKind.PlusPlusToken ||
                    node.operator === ts.SyntaxKind.MinusMinusToken;
            case ts.SyntaxKind.NewExpression:
                return this.options.allowNew;
            case ts.SyntaxKind.TaggedTemplateExpression:
                return this.options.allowTaggedTemplate;
            case ts.SyntaxKind.ParenthesizedExpression:
                return this.isUsed(node.expression);
            case ts.SyntaxKind.BinaryExpression: {
                const { operatorToken: { kind: tokenKind }, right } = node;
                switch (tokenKind) {
                    case ts.SyntaxKind.AmpersandAmpersandToken:
                    case ts.SyntaxKind.BarBarToken:
                        return this.options.allowShortCircuit && this.isUsed(right);
                    case ts.SyntaxKind.CommaToken:
                        return this.isUsed(right);
                    default:
                        return tsutils_1.isAssignmentKind(tokenKind);
                }
            }
            case ts.SyntaxKind.ConditionalExpression: {
                if (!this.options.allowTernary)
                    return false;
                const { whenTrue, whenFalse } = node;
                const whenTrueUsed = this.isUsed(whenTrue);
                const whenFalseUsed = this.isUsed(whenFalse);
                if (whenTrueUsed === whenFalseUsed)
                    return whenTrueUsed;
                this.addFindingAtNode(whenFalseUsed ? whenTrue : whenFalse, FAIL_MESSAGE);
                return true;
            }
            default:
                return false;
        }
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
/**
 * A directive is an expression statement containing only a string literal.
 * Directives need to be located at the beginning of a function block and be consecutive.
 */
function isDirective(statement) {
    if (statement.expression.kind !== ts.SyntaxKind.StringLiteral)
        return false;
    const parent = statement.parent;
    if (!utils_1.hasDirectivePrologue(parent))
        return false;
    for (let i = parent.statements.indexOf(statement) - 1; i >= 0; --i)
        if (parent.statements[i].kind !== ts.SyntaxKind.ExpressionStatement ||
            parent.statements[i].expression.kind !== ts.SyntaxKind.StringLiteral)
            return false;
    return true;
}
/** `void 0` and `void(0)` are allowed to have no side effect. */
function isAllowedVoidExpression(expr) {
    if (expr.kind === ts.SyntaxKind.ParenthesizedExpression)
        expr = expr.expression;
    return tsutils_1.isNumericLiteral(expr) && expr.text === '0';
}
/** Allow `(0, eval)('foo')` */
function isIndirectEval(node) {
    return tsutils_1.isIdentifier(node.right) && node.right.text === 'eval' &&
        tsutils_1.isNumericLiteral(node.left) && node.left.text === '0' &&
        node.parent.kind === ts.SyntaxKind.ParenthesizedExpression &&
        node.parent.parent.kind === ts.SyntaxKind.CallExpression;
}
//# sourceMappingURL=no-unused-expression.js.map