"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const utils_1 = require("../utils");
const tsutils_1 = require("tsutils");
const debug = require("debug");
const log = debug('wotan:rule:no-useless-assertion');
const FAIL_MESSAGE = "This assertion is unnecesary as it doesn't change the type of the expression.";
const FAIL_DEFINITE_ASSIGNMENT = 'This assertion is unnecessary as it has no effect on this declaration.';
let Rule = class Rule extends ymir_1.TypedRule {
    constructor() {
        super(...arguments);
        this.strictNullChecks = tsutils_1.isStrictCompilerOptionEnabled(this.context.compilerOptions, 'strictNullChecks');
    }
    apply() {
        for (const node of this.context.getFlatAst()) {
            switch (node.kind) {
                case ts.SyntaxKind.NonNullExpression:
                    this.checkNonNullAssertion(node);
                    break;
                case ts.SyntaxKind.AsExpression:
                case ts.SyntaxKind.TypeAssertionExpression:
                    this.checkTypeAssertion(node);
                    break;
                case ts.SyntaxKind.VariableDeclaration:
                    this.checkDefiniteAssignmentAssertion(node);
                    break;
                case ts.SyntaxKind.PropertyDeclaration:
                    this.checkDefiniteAssignmentAssertionProperty(node);
            }
        }
    }
    checkDefiniteAssignmentAssertion(node) {
        // compiler already emits an error for definite assignment assertions on ambient or initialized variables
        if (node.exclamationToken !== undefined &&
            node.initializer === undefined && (!tsutils_1.isStrictCompilerOptionEnabled(this.context.compilerOptions, 'strictNullChecks') ||
            getNullableFlags(this.checker.getTypeAtLocation(node.name), true) & ts.TypeFlags.Undefined // type does not allow undefined
        ))
            this.addFinding(node.exclamationToken.end - 1, node.exclamationToken.end, FAIL_DEFINITE_ASSIGNMENT, ymir_1.Replacement.delete(node.exclamationToken.pos, node.exclamationToken.end));
    }
    checkDefiniteAssignmentAssertionProperty(node) {
        // compiler emits an error for definite assignment assertions on ambient, initialized or abstract properties
        if (node.exclamationToken !== undefined &&
            node.initializer === undefined &&
            !tsutils_1.hasModifier(node.modifiers, ts.SyntaxKind.AbstractKeyword) && (node.name.kind !== ts.SyntaxKind.Identifier || // properties with string or computed name are not checked
            !tsutils_1.isStrictCompilerOptionEnabled(this.context.compilerOptions, 'strictPropertyInitialization') ||
            getNullableFlags(this.checker.getTypeAtLocation(node), true) & ts.TypeFlags.Undefined // type does not allow undefined
        ))
            this.addFinding(node.exclamationToken.end - 1, node.exclamationToken.end, FAIL_DEFINITE_ASSIGNMENT, ymir_1.Replacement.delete(node.exclamationToken.pos, node.exclamationToken.end));
    }
    checkNonNullAssertion(node) {
        let message = FAIL_MESSAGE;
        if (this.strictNullChecks) {
            const originalType = this.checker.getTypeAtLocation(node.expression);
            const flags = getNullableFlags(this.checker.getBaseConstraintOfType(originalType) || originalType);
            if (flags !== 0) { // type is nullable
                const contextualType = this.getSafeContextualType(node);
                if (contextualType === undefined || (flags & ~getNullableFlags(contextualType, true)))
                    return;
                message = `This assertion is unnecessary as the receiver accepts ${formatNullableFlags(flags)} values.`;
            }
            if (maybeUsedBeforeBeingAssigned(node.expression, originalType, this.checker)) {
                log('Identifier %s could be used before being assigned', node.expression.text);
                return;
            }
        }
        this.addFinding(node.end - 1, node.end, message, ymir_1.Replacement.delete(node.expression.end, node.end));
    }
    checkTypeAssertion(node) {
        if (tsutils_1.isConstAssertion(node)) {
            if (tsutils_1.isInConstContext(node))
                this.reportUselessTypeAssertion(node, 'This assertion is unnecessary as it is already in a const context.');
            return;
        }
        let targetType = this.checker.getTypeFromTypeNode(node.type);
        if ((targetType.flags & ts.TypeFlags.Literal) !== 0 && !tsutils_1.isInConstContext(node) || // allow "foo" as "foo" to avoid widening
            tsutils_1.isTupleType(tsutils_1.isTypeReference(targetType) ? targetType.target : targetType))
            return;
        let sourceType = this.checker.getTypeAtLocation(node.expression);
        if ((targetType.flags & (ts.TypeFlags.TypeVariable | ts.TypeFlags.Instantiable)) === 0) {
            targetType = this.checker.getBaseConstraintOfType(targetType) || targetType;
            sourceType = this.checker.getBaseConstraintOfType(sourceType) || sourceType;
        }
        let message = FAIL_MESSAGE;
        if (!utils_1.typesAreEqual(sourceType, targetType, this.checker)) {
            const contextualType = this.getSafeContextualType(node);
            // TODO use assignability check once it's exposed from TypeChecker
            if (contextualType === undefined ||
                contextualType.flags & (ts.TypeFlags.TypeVariable | ts.TypeFlags.Instantiable) ||
                (contextualType.flags & (ts.TypeFlags.Any | ts.TypeFlags.Unknown)) === 0 &&
                    // contextual type is exactly the same
                    !utils_1.typesAreEqual(sourceType, contextualType, this.checker) &&
                    // contextual type is an optional parameter or similar
                    !utils_1.typesAreEqual(sourceType, tsutils_1.removeOptionalityFromType(this.checker, contextualType), this.checker))
                return;
            message = 'This assertion is unnecessary as the receiver accepts the original type of the expression.';
        }
        this.reportUselessTypeAssertion(node, message);
    }
    reportUselessTypeAssertion(node, message) {
        if (node.kind === ts.SyntaxKind.AsExpression) {
            this.addFinding(node.type.pos - 'as'.length, node.end, message, ymir_1.Replacement.delete(node.expression.end, node.end));
        }
        else {
            const start = node.getStart(this.sourceFile);
            const fix = [ymir_1.Replacement.delete(start, node.expression.getStart(this.sourceFile))];
            if (utils_1.expressionNeedsParensWhenReplacingNode(node.expression, node))
                fix.push(ymir_1.Replacement.append(start, '('), ymir_1.Replacement.append(node.end, ')'));
            this.addFinding(start, node.expression.pos, message, fix);
        }
    }
    /** Returns the contextual type if it is a position that does not contribute to control flow analysis. */
    getSafeContextualType(node) {
        const parent = node.parent;
        // If assertion is used as argument, check if the function accepts this expression without an assertion
        // TODO expand this to VariableLike initializers and return expressions where a type declaration exists
        switch (parent.kind) {
            case ts.SyntaxKind.CallExpression:
            case ts.SyntaxKind.NewExpression:
                if (node === parent.expression)
                    return;
                break;
            case ts.SyntaxKind.TemplateSpan: // TODO return 'any' for non-tagged template expressions
            case ts.SyntaxKind.JsxExpression:
                break;
            default:
                return;
        }
        return this.checker.getContextualType(node);
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles,
    ymir_1.typescriptOnly
], Rule);
exports.Rule = Rule;
function getNullableFlags(type, receiver) {
    let flags = 0;
    for (const t of tsutils_1.unionTypeParts(type))
        flags |= t.flags;
    return ((receiver && flags & (ts.TypeFlags.Any | ts.TypeFlags.Unknown)) ? -1 : flags) & (ts.TypeFlags.Null | ts.TypeFlags.Undefined);
}
function formatNullableFlags(flags) {
    switch (flags) {
        case ts.TypeFlags.Null: return "'null'";
        case ts.TypeFlags.Undefined: return "'undefined'";
        default: return "'null' and 'undefined'";
    }
}
/**
 * Determine if the assertion may be necessary to avoid a compile error for use before being assigned.
 * This function may yield some false negatives, but is needed until https://github.com/Microsoft/TypeScript/issues/20221 is implemented.
 *
 * The rules are as follows:
 * * strictNullChecks are enabled
 * * assertion is on an identifier
 * * identifier is a mutable variable
 * * no destructuring, parameters, catch binding, ambient declarations
 * * variable is not initialized
 * * variable has no definite assignment assertion (exclamation mark)
 * * variable has a type annotation
 * * declared type is equal to the type at the assertion
 * * declaration and assertion are in the same function scope
 *
 * We don't need to worry about strictPropertyInitialization errors, because they cannot be suppressed with a non-null assertion
 */
function maybeUsedBeforeBeingAssigned(node, type, checker) {
    if (node.kind !== ts.SyntaxKind.Identifier || getNullableFlags(type, true) & ts.TypeFlags.Undefined)
        return false;
    const symbol = checker.getSymbolAtLocation(node);
    const declaration = symbol.declarations[0];
    if (!tsutils_1.isVariableDeclaration(declaration) ||
        declaration.parent.kind !== ts.SyntaxKind.VariableDeclarationList ||
        declaration.initializer !== undefined ||
        declaration.exclamationToken !== undefined ||
        declaration.type === undefined ||
        declaration.parent.parent.kind === ts.SyntaxKind.VariableStatement &&
            tsutils_1.hasModifier(declaration.parent.parent.modifiers, ts.SyntaxKind.DeclareKeyword))
        return false;
    if (!utils_1.typesAreEqual(type, checker.getTypeFromTypeNode(declaration.type), checker))
        return false;
    const declaringFunctionScope = findupFunction(declaration.parent.parent.parent);
    let useFunctionScope = findupFunction(node.parent.parent);
    while (useFunctionScope !== declaringFunctionScope && isInlinedIife(useFunctionScope))
        // TypeScript inlines IIFEs, so we need to do as well
        useFunctionScope = findupFunction(useFunctionScope.parent.parent);
    return useFunctionScope === declaringFunctionScope;
}
/** Finds the nearest parent that has a function scope. */
function findupFunction(node) {
    while (!tsutils_1.isFunctionScopeBoundary(node) && node.kind !== ts.SyntaxKind.SourceFile)
        node = node.parent;
    return node;
}
function isInlinedIife(node) {
    return (tsutils_1.isFunctionExpression(node) || tsutils_1.isArrowFunction(node)) &&
        node.asteriskToken === undefined && // exclude generators
        !tsutils_1.hasModifier(node.modifiers, ts.SyntaxKind.AsyncKeyword) && // exclude async functions
        tsutils_1.getIIFE(node) !== undefined;
}
//# sourceMappingURL=no-useless-assertion.js.map