"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
let Rule = class Rule extends ymir_1.AbstractRule {
    apply() {
        for (const node of this.context.getFlatAst()) {
            switch (node.kind) {
                case ts.SyntaxKind.ArrayLiteralExpression:
                    if (tsutils_1.isReassignmentTarget(node))
                        this.checkArray(node, node.parent.kind === ts.SyntaxKind.BinaryExpression, isNonEmptyAssignmentTarget, simplifyAssignmentRestElement);
                    break;
                case ts.SyntaxKind.ObjectLiteralExpression:
                    if (tsutils_1.isReassignmentTarget(node))
                        this.checkObject(node.properties, node.parent.kind === ts.SyntaxKind.BinaryExpression, isAssignmentPropertyUsed, tsutils_1.isSpreadAssignment, node);
                    break;
                case ts.SyntaxKind.ArrayBindingPattern:
                    this.checkArray(node, shouldReportEmptyBindingPattern(node), isBindingElementUsed, simplifyBindingRestElement);
                    break;
                case ts.SyntaxKind.ObjectBindingPattern:
                    this.checkObject(node.elements, shouldReportEmptyBindingPattern(node), isBindingPropertyUsed, isBindingRestProperty, node);
            }
        }
    }
    checkObject(properties, reportEmpty, isUsed, isRestProperty, node) {
        if (properties.length === 0) {
            if (reportEmpty)
                this.addFindingAtNode(node, 'Object destructuring is not necessary as it contains no properties.');
            return;
        }
        if (isRestProperty(properties[properties.length - 1]))
            return; // don't report empty property destructuring if there is a rest property as it's used to exclude properties
        for (let i = 0;; ++i) {
            const isLast = i === properties.length - 1;
            const property = properties[i];
            if (!isUsed(property))
                this.addFindingAtNode(property, "Destructuring property is not necessary as it doesn't assign to a variable.", 
                // make sure to delete the next comma to avoid consecutive commas
                ymir_1.Replacement.delete(property.pos, isLast ? properties.end : properties[i + 1].pos));
            if (isLast)
                break;
        }
    }
    checkArray(node, reportEmpty, isUsed, trySimplifyRest) {
        const elements = node.elements;
        if (elements.length === 0) {
            if (reportEmpty)
                this.addFindingAtNode(node, 'Array destructuring is not necessary as it contains no elements.');
            return;
        }
        const last = elements[elements.length - 1];
        if (isUsed(last)) {
            const restDestructuring = trySimplifyRest(last);
            if (restDestructuring !== undefined)
                this.addFindingAtNode(last, 'Array destructuring within array rest is redundant.', [
                    ymir_1.Replacement.delete(last.pos, restDestructuring.elements.pos),
                    ymir_1.Replacement.delete(restDestructuring.elements.end, last.end),
                ]);
            return;
        }
        let i = elements.length - 1;
        while (true) {
            if (i === 0 || isUsed(elements[i - 1])) {
                const diff = elements.length - i;
                this.addFinding(elements[i].getStart(this.sourceFile), elements.end, diff === 1
                    ? "Destructuring element is not necessary as it doesn't assign to a variable."
                    : "Destructuring elements are not necessary as they don't assign to a variable.", ymir_1.Replacement.delete(elements[i].pos, elements.end));
                break;
            }
            --i;
        }
        for (i -= 2; i >= 0; --i) {
            const element = elements[i];
            if (element.kind !== ts.SyntaxKind.OmittedExpression && !isUsed(element))
                this.addFindingAtNode(element, "Destructuring element is not necessary as it doesn't assign to a variable.", ymir_1.Replacement.delete(element.getStart(this.sourceFile), element.end));
        }
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
function isBindingPropertyUsed(node) {
    return node.dotDotDotToken !== undefined || node.propertyName === undefined || isNonEmptyBindingTarget(node.name);
}
function isBindingElementUsed(node) {
    return node.kind !== ts.SyntaxKind.OmittedExpression && isNonEmptyBindingTarget(node.name);
}
function isNonEmptyBindingTarget(node) {
    return node.kind === ts.SyntaxKind.Identifier || node.elements.length !== 0;
}
function isAssignmentPropertyUsed(node) {
    return node.kind !== ts.SyntaxKind.PropertyAssignment || isNonEmptyAssignmentTarget(node.initializer);
}
function isNonEmptyAssignmentTarget(node) {
    switch (node.kind) {
        case ts.SyntaxKind.OmittedExpression: // only possible in array destructuring
            return false;
        case ts.SyntaxKind.SpreadElement: // only possible in array destructuring
            return isNonEmptyAssignmentTarget(node.expression);
        case ts.SyntaxKind.ObjectLiteralExpression:
            return node.properties.length !== 0;
        case ts.SyntaxKind.ArrayLiteralExpression:
            return node.elements.length !== 0;
        default:
            return true;
    }
}
function simplifyBindingRestElement(node) {
    return node.kind === ts.SyntaxKind.OmittedExpression ||
        node.dotDotDotToken === undefined ||
        node.name.kind !== ts.SyntaxKind.ArrayBindingPattern
        ? undefined
        : node.name;
}
function simplifyAssignmentRestElement(node) {
    return tsutils_1.isSpreadElement(node) && tsutils_1.isArrayLiteralExpression(node.expression) ? node.expression : undefined;
}
function isBindingRestProperty(node) {
    return node.dotDotDotToken !== undefined;
}
function shouldReportEmptyBindingPattern(node) {
    switch (node.parent.kind) {
        case ts.SyntaxKind.BindingElement:
        case ts.SyntaxKind.Parameter:
            return false;
        default:
            return true;
    }
}
//# sourceMappingURL=no-useless-destructuring.js.map