"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
let Rule = class Rule extends ymir_1.AbstractRule {
    apply() {
        const isJs = this.sourceFile.flags & ts.NodeFlags.JavaScriptFile;
        for (const node of this.context.getFlatAst()) {
            switch (node.kind) {
                case ts.SyntaxKind.VariableDeclaration:
                    if (node.initializer !== undefined &&
                        (node.parent.flags & ts.NodeFlags.Const) === 0 &&
                        isUndefined(node.initializer))
                        this.fail(node);
                    break;
                case ts.SyntaxKind.ObjectBindingPattern:
                    this.checkBindingPattern(node, getObjectPropertyName);
                    break;
                case ts.SyntaxKind.ArrayBindingPattern:
                    this.checkBindingPattern(node, getArrayPropertyName);
                    break;
                case ts.SyntaxKind.ObjectLiteralExpression:
                    if (tsutils_1.isReassignmentTarget(node))
                        this.checkObjectDestructuring(node);
                    break;
                case ts.SyntaxKind.ArrayLiteralExpression:
                    if (tsutils_1.isReassignmentTarget(node))
                        this.checkArrayDestructuring(node);
                    break;
                default:
                    if (!isJs && tsutils_1.isFunctionWithBody(node))
                        this.checkFunctionParameters(node.parameters);
            }
        }
    }
    checkObjectDestructuring(node) {
        if (this.context.compilerOptions === undefined || !tsutils_1.isStrictCompilerOptionEnabled(this.context.compilerOptions, 'strictNullChecks'))
            return;
        const checker = this.program.getTypeChecker();
        let type;
        for (const property of node.properties) {
            let name;
            let errorNode;
            switch (property.kind) {
                case ts.SyntaxKind.ShorthandPropertyAssignment:
                    if (property.objectAssignmentInitializer === undefined)
                        continue;
                    name = property.name;
                    errorNode = property.objectAssignmentInitializer;
                    break;
                case ts.SyntaxKind.PropertyAssignment:
                    if (!tsutils_1.isBinaryExpression(property.initializer))
                        continue;
                    name = property.name;
                    errorNode = property.initializer.right;
                    break;
                default:
                    continue;
            }
            // for compatibility with typescript@<3.5.0
            if (checker.getTypeOfAssignmentPattern !== undefined) {
                const properties = tsutils_1.getLateBoundPropertyNamesOfPropertyName(name, checker);
                if (!properties.known || properties.names.some(maybeUndefined))
                    continue;
            }
            else if (name.kind !== ts.SyntaxKind.Identifier ||
                symbolMaybeUndefined(checker, checker.getPropertySymbolOfDestructuringAssignment(name), node)) {
                continue;
            }
            this.addFindingAtNode(errorNode, "Unnecessary default value as this property is never 'undefined'.", ymir_1.Replacement.delete(tsutils_1.getChildOfKind(errorNode.parent, ts.SyntaxKind.EqualsToken, this.sourceFile).pos, errorNode.end));
        }
        function maybeUndefined({ symbolName }) {
            return symbolMaybeUndefined(checker, tsutils_1.getPropertyOfType(type || (type = checker.getApparentType(checker.getTypeOfAssignmentPattern(node))), symbolName), node);
        }
    }
    checkArrayDestructuring(node) {
        if (this.context.compilerOptions === undefined || !tsutils_1.isStrictCompilerOptionEnabled(this.context.compilerOptions, 'strictNullChecks'))
            return;
        const checker = this.program.getTypeChecker();
        if (checker.getTypeOfAssignmentPattern === undefined)
            return; // for compatibility with typescript@<3.5.0
        let type;
        for (let i = 0; i < node.elements.length; ++i) {
            const element = node.elements[i];
            if (!tsutils_1.isBinaryExpression(element))
                continue;
            if (type === undefined) {
                type = checker.getTypeOfAssignmentPattern(node);
                type = checker.getBaseConstraintOfType(type) || type;
            }
            if (symbolMaybeUndefined(checker, type.getProperty(String(i)), node))
                continue;
            this.addFindingAtNode(element.right, "Unnecessary default value as this element is never 'undefined'.", ymir_1.Replacement.delete(element.operatorToken.pos, element.end));
        }
    }
    checkBindingPattern(node, propNames) {
        if (this.context.compilerOptions === undefined || !tsutils_1.isStrictCompilerOptionEnabled(this.context.compilerOptions, 'strictNullChecks'))
            return;
        const checker = this.program.getTypeChecker();
        let type;
        for (let i = 0; i < node.elements.length; ++i) {
            const element = node.elements[i];
            if (element.kind === ts.SyntaxKind.OmittedExpression || element.initializer === undefined)
                continue;
            const lateBoundNames = propNames(element, i, checker);
            if (!lateBoundNames.known || lateBoundNames.names.some(maybeUndefined))
                continue;
            // TODO we currently cannot autofix this case: it's possible to use a default value that's not assignable to the
            // destructured type. The type of the variable then includes the type of the initializer as well.
            // Removing the initializer might also remove its type from the union type causing type errors elsewhere.
            this.addFindingAtNode(element.initializer, "Unnecessary default value as this property is never 'undefined'.");
        }
        function maybeUndefined({ symbolName }) {
            return symbolMaybeUndefined(checker, tsutils_1.getPropertyOfType(type || (type = checker.getApparentType(checker.getTypeAtLocation(node))), symbolName), node);
        }
    }
    checkFunctionParameters(parameters) {
        let allOptional = true;
        for (let i = parameters.length - 1; i >= 0; --i) {
            const parameter = parameters[i];
            if (parameter.initializer !== undefined) {
                if (isUndefined(parameter.initializer))
                    this.fail(parameter, allOptional);
            }
            else if (parameter.questionToken === undefined && parameter.dotDotDotToken === undefined) {
                allOptional = false;
            }
        }
    }
    fail(node, makeOptional) {
        const fix = [ymir_1.Replacement.delete(tsutils_1.getChildOfKind(node, ts.SyntaxKind.EqualsToken, this.sourceFile).pos, node.end)];
        let message = "Unnecessary initialization with 'undefined'.";
        if (makeOptional) {
            fix.push(ymir_1.Replacement.append(node.name.end, '?'));
            const removeUndefined = this.removeUndefinedFromType(node.type);
            if (removeUndefined !== undefined)
                fix.push(removeUndefined);
            message += ' Use an optional parameter instead.';
        }
        this.addFinding(node.end - 'undefined'.length, node.end, message, fix);
    }
    removeUndefinedFromType(type) {
        if (type === undefined || !tsutils_1.isUnionTypeNode(type))
            return;
        for (let i = 0; i < type.types.length; ++i) {
            const t = type.types[i];
            if (t.kind === ts.SyntaxKind.UndefinedKeyword) {
                const bar = (i === 0 ? tsutils_1.getNextToken : tsutils_1.getPreviousToken)(t, this.sourceFile);
                return ymir_1.Replacement.delete(Math.min(t.pos, bar.pos), Math.max(t.end, bar.end));
            }
        }
        return;
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
function getObjectPropertyName(property, _i, checker) {
    if (property.propertyName === undefined)
        return {
            known: true,
            names: [{ displayName: property.name.text, symbolName: property.name.escapedText }],
        };
    return tsutils_1.getLateBoundPropertyNamesOfPropertyName(property.propertyName, checker);
}
function getArrayPropertyName(_, i) {
    const name = String(i);
    return { known: true, names: [{ displayName: name, symbolName: name }] };
}
function symbolMaybeUndefined(checker, symbol, node) {
    if (symbol === undefined)
        return true;
    if (symbol.flags & ts.SymbolFlags.Optional)
        return true;
    return typeMaybeUndefined(checker, checker.getTypeOfSymbolAtLocation(symbol, node));
}
function typeMaybeUndefined(checker, type) {
    if (tsutils_1.isInstantiableType(type)) {
        const constraint = checker.getBaseConstraintOfType(type);
        if (constraint === undefined)
            return true;
        type = constraint;
    }
    if (tsutils_1.isUnionType(type))
        return type.types.some((t) => typeMaybeUndefined(checker, t));
    return (type.flags & (ts.TypeFlags.Undefined | ts.TypeFlags.Any | ts.TypeFlags.Unknown)) !== 0;
}
function isUndefined(node) {
    return tsutils_1.isIdentifier(node) && node.originalKeywordKind === ts.SyntaxKind.UndefinedKeyword;
}
//# sourceMappingURL=no-useless-initializer.js.map