"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
const utils_1 = require("../utils");
const primitiveFlags = ts.TypeFlags.BigIntLike | ts.TypeFlags.BooleanLike | ts.TypeFlags.NumberLike | ts.TypeFlags.StringLike |
    ts.TypeFlags.ESSymbolLike | ts.TypeFlags.Undefined | ts.TypeFlags.Void;
const predicates = {
    object: {
        nullable: true,
        check(type) {
            return tsutils_1.isEmptyObjectType(type) ? undefined : (type.flags & primitiveFlags) === 0 && !isTypeofFunction(type);
        },
    },
    number: {
        nullable: false,
        check: checkFlags(ts.TypeFlags.NumberLike),
    },
    bigint: {
        nullable: false,
        check: checkFlags(ts.TypeFlags.BigIntLike),
    },
    string: {
        nullable: false,
        check: checkFlags(ts.TypeFlags.StringLike),
    },
    boolean: {
        nullable: false,
        check: checkFlags(ts.TypeFlags.BooleanLike),
    },
    symbol: {
        nullable: false,
        check: checkFlags(ts.TypeFlags.ESSymbolLike),
    },
    function: {
        nullable: false,
        check: isTypeofFunction,
    },
    null: {
        nullable: true,
        check: checkFlags(ts.TypeFlags.Null),
    },
    undefined: {
        nullable: true,
        check: checkFlags(ts.TypeFlags.Undefined | ts.TypeFlags.Void),
    },
    nullOrUndefined: {
        nullable: true,
        check: checkFlags(ts.TypeFlags.Null | ts.TypeFlags.Undefined | ts.TypeFlags.Void),
    },
};
let Rule = class Rule extends ymir_1.TypedRule {
    constructor() {
        super(...arguments);
        this.strictNullChecks = tsutils_1.isStrictCompilerOptionEnabled(this.context.compilerOptions, 'strictNullChecks');
    }
    apply() {
        for (const node of this.context.getFlatAst()) {
            switch (node.kind) {
                case ts.SyntaxKind.IfStatement:
                    this.checkCondition(node.expression);
                    break;
                case ts.SyntaxKind.ConditionalExpression:
                    this.checkCondition(node.condition);
                    break;
                case ts.SyntaxKind.WhileStatement:
                case ts.SyntaxKind.DoStatement:
                    if (node.expression.kind !== ts.SyntaxKind.TrueKeyword)
                        this.checkCondition(node.expression);
                    break;
                case ts.SyntaxKind.ForStatement: {
                    const { condition } = node;
                    if (condition !== undefined && condition.kind !== ts.SyntaxKind.TrueKeyword)
                        this.checkCondition(condition);
                    break;
                }
                case ts.SyntaxKind.SwitchStatement:
                    this.checkSwitch(node);
                    break;
                case ts.SyntaxKind.BinaryExpression:
                    if (isLogicalOperator(node.operatorToken.kind)) {
                        this.checkCondition(node.left);
                    }
                    else if (isEqualityOrInOperator(node.operatorToken.kind)) {
                        this.checkNode(node);
                    }
                    break;
                case ts.SyntaxKind.PrefixUnaryExpression:
                    if (node.operator === ts.SyntaxKind.ExclamationToken)
                        this.checkCondition(node.operand);
            }
        }
    }
    checkSwitch(node) {
        for (const clause of node.caseBlock.clauses) {
            if (clause.kind === ts.SyntaxKind.DefaultClause)
                continue;
            this.maybeFail(clause, this.isConstantComparison(node.expression, clause.expression, ts.SyntaxKind.EqualsEqualsEqualsToken));
        }
    }
    checkCondition(node) {
        return this.maybeFail(node, this.isTruthyFalsy(node, true));
    }
    checkNode(node) {
        return this.maybeFail(node, this.isTruthyFalsy(node, false));
    }
    maybeFail(node, result) {
        if (result !== undefined)
            return this.addFindingAtNode(node, result ? 'Expression is always truthy.' : 'Expression is always falsy.');
    }
    isTruthyFalsy(node, nested) {
        if (tsutils_1.isBinaryExpression(node)) {
            if (isEqualityOperator(node.operatorToken.kind)) // TODO check <, >, <=, >= with literal types
                return nested ? undefined : this.isConstantComparison(node.left, node.right, node.operatorToken.kind);
            if (node.operatorToken.kind === ts.SyntaxKind.InKeyword)
                return nested ? undefined : this.isPropertyPresent(node.right, node.left);
        }
        else if (tsutils_1.isPrefixUnaryExpression(node) && node.operator === ts.SyntaxKind.ExclamationToken) {
            return;
        }
        else if (tsutils_1.isParenthesizedExpression(node)) {
            return this.isTruthyFalsy(node.expression, true);
        }
        // in non-strictNullChecks mode we can only detect if a type is definitely falsy
        return this.executePredicate(this.getTypeOfExpression(node), this.strictNullChecks ? truthyFalsy : falsy);
    }
    isConstantComparison(left, right, operator) {
        left = tsutils_1.unwrapParentheses(left);
        right = tsutils_1.unwrapParentheses(right);
        const equals = {
            negated: operator === ts.SyntaxKind.ExclamationEqualsEqualsToken || operator === ts.SyntaxKind.ExclamationEqualsToken,
            strict: operator === ts.SyntaxKind.ExclamationEqualsEqualsToken || operator === ts.SyntaxKind.EqualsEqualsEqualsToken,
        };
        let result = this.checkEquals(left, right, equals);
        if (result === undefined)
            result = this.checkEquals(right, left, equals);
        return result;
    }
    checkEquals(left, right, equals) {
        let predicate;
        if (tsutils_1.isTypeOfExpression(left)) {
            left = tsutils_1.unwrapParentheses(left.expression);
            if (right.kind === ts.SyntaxKind.NullKeyword || isUndefined(right))
                return equals.negated;
            let literal;
            if (tsutils_1.isTextualLiteral(right)) {
                literal = right.text;
            }
            else {
                let type = this.getTypeOfExpression(right);
                type = this.checker.getBaseConstraintOfType(type) || type;
                if ((type.flags & ts.TypeFlags.StringLiteral) === 0)
                    return;
                literal = type.value;
            }
            switch (literal) {
                default:
                    return equals.negated;
                case 'number':
                case 'string':
                case 'boolean':
                case 'symbol':
                case 'object':
                case 'function':
                case 'undefined':
                case 'bigint':
                    predicate = predicates[literal];
            }
        }
        else if (right.kind === ts.SyntaxKind.NullKeyword) {
            predicate = equals.strict ? predicates.null : predicates.nullOrUndefined;
        }
        else if (isUndefined(right)) {
            predicate = equals.strict ? predicates.undefined : predicates.nullOrUndefined;
        }
        else if (this.strictNullChecks) {
            const leftLiteral = this.getPrimitiveLiteral(left);
            return leftLiteral !== undefined && leftLiteral === this.getPrimitiveLiteral(right) ? !equals.negated : undefined;
        }
        else {
            return;
        }
        return this.nullAwarePredicate(this.getTypeOfExpression(left), predicate);
    }
    getPrimitiveLiteral(node) {
        // TODO reuse some logic from 'no-duplicate-case' to compute prefix unary expressions
        let type = this.getTypeOfExpression(node);
        type = this.checker.getBaseConstraintOfType(type) || type;
        for (const t of tsutils_1.intersectionTypeParts(type)) {
            if (tsutils_1.isLiteralType(t))
                return typeof t.value === 'object' ? utils_1.formatPseudoBigInt(t.value) : String(t.value);
            if (t.flags & ts.TypeFlags.BooleanLiteral)
                return t.intrinsicName;
            if (t.flags & ts.TypeFlags.Undefined)
                return 'undefined';
            if (t.flags & ts.TypeFlags.Null)
                return 'null';
        }
        return;
    }
    nullAwarePredicate(type, predicate) {
        if (!this.strictNullChecks && predicate.nullable)
            return;
        const result = this.executePredicate(type, 
        // empty object type can contain anything but `null | undefined`
        // TODO use assignability check to avoid false positives
        predicate.nullable ? predicate.check : (t) => tsutils_1.isEmptyObjectType(t) ? undefined : predicate.check(t));
        return result && !this.strictNullChecks
            ? undefined
            : result;
    }
    executePredicate(type, predicate) {
        let result;
        for (let t of tsutils_1.unionTypeParts(type)) {
            if (t.flags & (ts.TypeFlags.TypeVariable | ts.TypeFlags.Instantiable)) {
                const constraint = this.checker.getBaseConstraintOfType(t);
                if (constraint === undefined)
                    return;
                t = constraint;
            }
            if (t.flags & (ts.TypeFlags.Any | ts.TypeFlags.Never | ts.TypeFlags.Unknown))
                return;
            switch (tsutils_1.isIntersectionType(t) ? this.matchIntersectionType(t, predicate) : predicate(t)) {
                case true:
                    if (result === false)
                        return;
                    result = true;
                    break;
                case false:
                    if (result === true)
                        return;
                    result = false;
                    break;
                default:
                    return;
            }
        }
        return result;
    }
    /**
     * If one of the types in the intersection matches the precicate, this function returns true.
     * Otherwise if any of the types results in undefined, this function returns undefined.
     * If every type results in false, it returns false.
     */
    matchIntersectionType(type, predicate) {
        let result = false;
        for (const t of type.types) {
            switch (this.executePredicate(t, predicate)) {
                case true:
                    return true;
                case undefined:
                    result = undefined;
            }
        }
        return result;
    }
    getTypeOfExpression(node) {
        const type = this.checker.getTypeAtLocation(node);
        if (!this.strictNullChecks)
            return type;
        if (tsutils_1.unionTypeParts(type).some((t) => (t.flags & ts.TypeFlags.Undefined) !== 0))
            return type;
        if (!tsutils_1.isPropertyAccessExpression(node) && !tsutils_1.isElementAccessExpression(node))
            return type;
        const objectType = this.checker.getApparentType(this.checker.getTypeAtLocation(node.expression));
        if (objectType.getStringIndexType() === undefined && objectType.getNumberIndexType() === undefined)
            return type;
        if (tsutils_1.isPropertyAccessExpression(node)) {
            if (objectType.getProperty(node.name.text) !== undefined)
                return type;
        }
        else {
            const names = tsutils_1.getLateBoundPropertyNames(node.argumentExpression, this.checker);
            if (names.known && names.names.every(({ symbolName }) => tsutils_1.getPropertyOfType(objectType, symbolName) !== undefined))
                return type;
        }
        return this.checker.getNullableType(type, ts.TypeFlags.Undefined);
    }
    isPropertyPresent(node, name) {
        if (!this.strictNullChecks)
            return;
        const names = tsutils_1.getLateBoundPropertyNames(name, this.checker); // TODO lateBoundPropertyNames should also be aware of index signatures
        if (!names.known)
            return;
        const types = tsutils_1.unionTypeParts(this.checker.getApparentType(this.getTypeOfExpression(tsutils_1.unwrapParentheses(node))));
        for (const { symbolName } of names.names) {
            // we check every union type member separately because symbols lose optionality when accessed through union types
            for (const type of types) {
                const symbol = tsutils_1.getPropertyOfType(type, symbolName);
                if (symbol === undefined || symbol.flags & ts.SymbolFlags.Optional)
                    return; // it might be present at runtime, so we don't return false
            }
        }
        return true;
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
function isUndefined(node) {
    return tsutils_1.isIdentifier(node) && node.originalKeywordKind === ts.SyntaxKind.UndefinedKeyword;
}
function falsy(type) {
    return tsutils_1.isFalsyType(type) ? false : undefined;
}
function truthyFalsy(type) {
    if (type.flags & ts.TypeFlags.PossiblyFalsy) {
        if (tsutils_1.isFalsyType(type))
            return false;
        return type.flags & ts.TypeFlags.Literal ? true : undefined;
    }
    // TODO use assignability check
    return tsutils_1.isEmptyObjectType(type) ? undefined : true;
}
function isLogicalOperator(kind) {
    return kind === ts.SyntaxKind.AmpersandAmpersandToken || kind === ts.SyntaxKind.BarBarToken;
}
function isEqualityOrInOperator(kind) {
    return isEqualityOperator(kind) || kind === ts.SyntaxKind.InKeyword;
}
function isEqualityOperator(kind) {
    switch (kind) {
        case ts.SyntaxKind.ExclamationEqualsEqualsToken:
        case ts.SyntaxKind.ExclamationEqualsToken:
        case ts.SyntaxKind.EqualsEqualsEqualsToken:
        case ts.SyntaxKind.EqualsEqualsToken:
            return true;
        default:
            return false;
    }
}
function checkFlags(flags) {
    return (type) => (type.flags & flags) !== 0;
}
function isTypeofFunction(type) {
    if (type.getCallSignatures().length !== 0 || type.getConstructSignatures().length !== 0)
        return true;
    // check if this could be the global `Function` type
    return type.symbol !== undefined && type.symbol.escapedName === 'Function' &&
        hasPropertyOfKind(type, 'apply', ts.SymbolFlags.Method) &&
        hasPropertyOfKind(type, 'arguments', ts.SymbolFlags.Property) &&
        hasPropertyOfKind(type, 'bind', ts.SymbolFlags.Method) &&
        hasPropertyOfKind(type, 'call', ts.SymbolFlags.Method) &&
        hasPropertyOfKind(type, 'caller', ts.SymbolFlags.Property) &&
        hasPropertyOfKind(type, 'length', ts.SymbolFlags.Property) &&
        hasPropertyOfKind(type, 'prototype', ts.SymbolFlags.Property);
}
function hasPropertyOfKind(type, name, flag) {
    const property = type.getProperty(name);
    return property !== undefined && (property.flags & flag) !== 0;
}
//# sourceMappingURL=no-useless-predicate.js.map