"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ts = require("typescript");
const ymir_1 = require("@fimbul/ymir");
const tsutils_1 = require("tsutils");
const MESSAGE = 'Using the spread operator here is not necessary.';
let Rule = class Rule extends ymir_1.AbstractRule {
    apply() {
        const re = /\.{3}\s*[/[{]/g;
        let wrappedAst;
        for (let match = re.exec(this.sourceFile.text); match !== null; match = re.exec(this.sourceFile.text)) {
            const { node } = tsutils_1.getWrappedNodeAtPosition(wrappedAst || (wrappedAst = this.context.getWrappedAst()), match.index);
            switch (node.kind) {
                case ts.SyntaxKind.SpreadElement:
                    if (node.expression.pos - 3 === match.index)
                        this.checkSpreadElement(node);
                    break;
                case ts.SyntaxKind.SpreadAssignment:
                    if (node.expression.pos - 3 === match.index)
                        this.checkSpreadAssignment(node);
                    break;
                case ts.SyntaxKind.JsxSpreadAttribute:
                    if (node.expression.pos - 3 === match.index)
                        this.checkJsxSpreadAttribute(node);
            }
        }
    }
    checkSpreadElement(node) {
        if (!tsutils_1.isArrayLiteralExpression(node.expression) || node.expression.elements.some(tsutils_1.isOmittedExpression) && !tsutils_1.isReassignmentTarget(node))
            return;
        this.addFindingAtNode(node, MESSAGE, removeUselessSpread(node, node.expression.elements));
    }
    checkSpreadAssignment(node) {
        if (!tsutils_1.isObjectLiteralExpression(node.expression))
            return;
        this.addFindingAtNode(node, MESSAGE, removeUselessSpread(node, node.expression.properties));
    }
    checkJsxSpreadAttribute(node) {
        if (!tsutils_1.isObjectLiteralExpression(node.expression) || !canConvertObjectSpreadToJsx(node.expression))
            return;
        this.addFindingAtNode(node, MESSAGE, removeUselessJsxSpreadAttribute(node, node.expression.properties));
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
function canConvertObjectSpreadToJsx(node) {
    for (const property of node.properties) {
        switch (property.kind) {
            case ts.SyntaxKind.SpreadAssignment:
            case ts.SyntaxKind.ShorthandPropertyAssignment:
                break;
            case ts.SyntaxKind.PropertyAssignment:
                const staticName = tsutils_1.getPropertyName(property.name);
                if (staticName === undefined || !tsutils_1.isValidJsxIdentifier(staticName))
                    return false;
                break;
            default:
                // TODO do something special for MethodDeclaration?
                // exclude accessors
                return false;
        }
    }
    return true;
}
function removeUselessJsxSpreadAttribute(node, properties) {
    let prevEnd = node.pos;
    const fix = [];
    for (const property of properties) {
        fix.push(ymir_1.Replacement.replace(prevEnd, property.pos, ' ')); // use space instead of comma as separator
        switch (property.kind) {
            case ts.SyntaxKind.SpreadAssignment:
                fix.push(ymir_1.Replacement.append(property.expression.pos - 3, '{'), ymir_1.Replacement.append(property.end, '}'));
                break;
            case ts.SyntaxKind.PropertyAssignment:
                fix.push(ymir_1.Replacement.replace(property.getStart(), property.initializer.pos, `${tsutils_1.getPropertyName(property.name)}={`), ymir_1.Replacement.append(property.end, '}'));
                break;
            default:
                fix.push(ymir_1.Replacement.append(property.end, `={${property.name.text}}`));
        }
        prevEnd = property.end;
    }
    fix.push(ymir_1.Replacement.delete(prevEnd, node.end));
    return fix;
}
function removeUselessSpread(node, elements) {
    if (elements.length === 0) {
        // handle special case of empty object or array to remove trailing comma
        const parent = node.parent;
        const containingList = parent.kind === ts.SyntaxKind.ArrayLiteralExpression
            ? parent.elements
            : parent.kind === ts.SyntaxKind.ObjectLiteralExpression
                ? parent.properties
                : parent.arguments;
        const index = containingList.indexOf(node);
        return ymir_1.Replacement.delete(node.pos, index === containingList.length - 1 ? containingList.end : containingList[index + 1].pos);
    }
    return [
        ymir_1.Replacement.delete(node.expression.pos - 3, elements.pos),
        ymir_1.Replacement.delete(elements[elements.length - 1].end, node.expression.end),
    ];
}
//# sourceMappingURL=no-useless-spread.js.map