"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const tsutils_1 = require("tsutils");
const ts = require("typescript");
const utils_1 = require("../utils");
let Rule = class Rule extends ymir_1.AbstractRule {
    apply() {
        for (const node of utils_1.tryStatements(this.context)) {
            const { tryBlock, catchClause, finallyBlock } = node;
            const start = getStart(node);
            const hasFinally = finallyBlock !== undefined && finallyBlock.statements.length !== 0;
            if (tryBlock.statements.length === 0) {
                this.addFinding(start, node.end, "'try' statement is unnecessary because the 'try' block is empty.", hasFinally ? deleteStatementLeavingBlock(node, finallyBlock, this.sourceFile) : deleteStatement(node));
            }
            else if (catchClause !== undefined && isRethrow(catchClause)) {
                // reminder for myself: empty catch clause can be used to simply ignore errors and is never redundant
                this.addFinding(start, node.end, `${hasFinally ? "'catch' clause" : "'try' statement"} is unnecessary because the 'catch' clause only rethrows the error.`, hasFinally
                    ? ymir_1.Replacement.delete(catchClause.getStart(this.sourceFile), finallyBlock.pos - 'finally'.length)
                    : deleteStatementLeavingBlock(node, tryBlock, this.sourceFile));
            }
            else if (finallyBlock !== undefined && finallyBlock.statements.length === 0) {
                if (catchClause === undefined) {
                    this.addFinding(start, node.end, "'try' statement is unnecessary because the 'finally' block is empty.", deleteStatementLeavingBlock(node, tryBlock, this.sourceFile));
                }
                else {
                    this.addFinding(finallyBlock.pos - 'finally'.length, node.end, "Empty 'finally' clause is unnecessary.", ymir_1.Replacement.delete(catchClause.end, finallyBlock.end));
                }
            }
        }
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
function deleteStatement(node) {
    return tsutils_1.isInSingleStatementContext(node)
        ? ymir_1.Replacement.replace(getStart(node), node.end, '{}')
        : ymir_1.Replacement.delete(node.pos, node.end);
}
function deleteStatementLeavingBlock(node, block, sourceFile) {
    const start = getStart(node);
    return canStripBlockOnRemove(node, block)
        ? [
            ymir_1.Replacement.delete(start, block.statements[0].getStart(sourceFile)),
            ymir_1.Replacement.delete(block.statements[block.statements.length - 1].end, node.end),
        ]
        : [
            ymir_1.Replacement.delete(start, block.statements.pos - 1),
            ymir_1.Replacement.delete(block.end, node.end),
        ];
}
function canStripBlockOnRemove(node, block) {
    return tsutils_1.isInSingleStatementContext(node)
        ? block.statements.length === 1 && !tsutils_1.isBlockScopedDeclarationStatement(block.statements[0])
        : !block.statements.some(tsutils_1.isBlockScopedDeclarationStatement);
}
function getStart(node) {
    return node.tryBlock.pos - 'try'.length;
}
function isRethrow(node) {
    return node.variableDeclaration !== undefined &&
        node.variableDeclaration.name.kind === ts.SyntaxKind.Identifier &&
        node.block.statements.length === 1 &&
        throwsVariable(node.block.statements[0], node.variableDeclaration.name);
}
function throwsVariable(statement, name) {
    if (!tsutils_1.isThrowStatement(statement) || statement.expression === undefined)
        return false;
    const expression = tsutils_1.unwrapParentheses(statement.expression);
    return tsutils_1.isIdentifier(expression) && expression.escapedText === name.escapedText;
}
//# sourceMappingURL=no-useless-try-catch.js.map