"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ts = require("typescript");
const ymir_1 = require("@fimbul/ymir");
const tsutils_1 = require("tsutils");
var Mode;
(function (Mode) {
    Mode[Mode["Never"] = 0] = "Never";
    Mode[Mode["WhenPossible"] = 1] = "WhenPossible";
    Mode[Mode["Consistent"] = 2] = "Consistent";
})(Mode = exports.Mode || (exports.Mode = {}));
const MESSAGES = {
    [Mode.Never]: 'Parameter properties have been disallowed.',
    [Mode.WhenPossible]: 'Use parameter properties when possible.',
    [Mode.Consistent]: {
        canBeParamPropsOnly: 'All parameters can be parameter properties.',
        cannotBeParamPropsOnly: 'Only use parameter properties if all parameters can be parameter properties.',
    },
};
let Rule = class Rule extends ymir_1.ConfigurableRule {
    parseOptions(options) {
        if (options) {
            switch (options.mode) {
                case 'consistent':
                    return { mode: Mode.Consistent };
                case 'never':
                    return { mode: Mode.Never };
            }
        }
        return { mode: Mode.WhenPossible };
    }
    apply() {
        for (const node of this.context.getFlatAst()) {
            if (tsutils_1.isClassDeclaration(node) || tsutils_1.isClassExpression(node)) {
                const construct = node.members.find((member) => tsutils_1.isConstructorDeclaration(member) && member.body !== undefined);
                if (construct)
                    this.checkConstructorParameters(construct);
            }
        }
    }
    checkConstructorParameters(construct) {
        switch (this.options.mode) {
            case Mode.Never:
                for (const param of construct.parameters.filter(tsutils_1.isParameterProperty))
                    this.addFindingAtNode(param, MESSAGES[Mode.Never], getFixerForDisallowedParameterProp(construct, param, tsutils_1.getLineBreakStyle(this.context.sourceFile)));
                break;
            case Mode.Consistent:
                if (construct.parameters.every(tsutils_1.isParameterProperty))
                    return;
                const allPropsCanBeParamProps = construct.parameters
                    .filter((param) => !tsutils_1.isParameterProperty(param))
                    .every((param) => canBeParameterProperty(param, construct));
                if (!allPropsCanBeParamProps && construct.parameters.some(tsutils_1.isParameterProperty)) {
                    for (const param of construct.parameters.filter(tsutils_1.isParameterProperty))
                        this.addFinding(param.pos, param.end, MESSAGES[Mode.Consistent].cannotBeParamPropsOnly, getFixerForDisallowedParameterProp(construct, param, tsutils_1.getLineBreakStyle(this.context.sourceFile)));
                }
                else if (allPropsCanBeParamProps && !construct.parameters.every(tsutils_1.isParameterProperty)) {
                    for (const param of construct.parameters.filter((p) => !tsutils_1.isParameterProperty(p)))
                        this.addFinding(param.pos, param.end, MESSAGES[Mode.Consistent].canBeParamPropsOnly, getFixerForLonghandProp(param, construct));
                }
                break;
            default:
                for (const param of construct.parameters)
                    if (!tsutils_1.isParameterProperty(param) && canBeParameterProperty(param, construct))
                        this.addFindingAtNode(param, MESSAGES[Mode.WhenPossible], getFixerForLonghandProp(param, construct));
        }
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles,
    ymir_1.typescriptOnly
], Rule);
exports.Rule = Rule;
function canBeParameterProperty(param, construct) {
    if (!tsutils_1.isIdentifier(param.name) || param.dotDotDotToken)
        return false;
    const member = construct.parent.members.find((mem) => isPropertyMatchForParam(mem, param));
    if (member && !member.decorators) {
        for (const stmt of statementsMinusDirectivesAndSuper(construct)) {
            if (isStatementWithPossibleSideEffects(stmt, param))
                return false;
            if (isSimpleParamToPropAssignment(stmt, param))
                return true;
        }
    }
    return false;
}
function getFixerForDisallowedParameterProp(construct, param, lineBreak) {
    return [
        /* Add assignment after directives and super() call (if they exists) */
        ymir_1.Replacement.append(getEndOfDirectivesAndSuper(construct), `${lineBreak}this.${param.name.getText()} = ${param.name.getText()};`),
        /* Add property to class */
        ymir_1.Replacement.append(construct.getStart(), getPropertyTextFromParam(param) + lineBreak),
        /* Finally, delete modifiers from the parameter prop */
        ymir_1.Replacement.delete(param.modifiers.pos, param.modifiers.end),
    ];
}
function getFixerForLonghandProp(param, construct) {
    /* Class member and assignment to be removed */
    const member = construct
        .parent.members.filter(tsutils_1.isPropertyDeclaration)
        .find((prop) => tsutils_1.getPropertyName(prop.name) === param.name.getText());
    const assignment = construct
        .body.statements.filter(tsutils_1.isExpressionStatement)
        .find((stmt) => isSimpleParamToPropAssignment(stmt, param));
    return [
        ymir_1.Replacement.delete(member.pos, member.end),
        ymir_1.Replacement.delete(assignment.pos, assignment.end),
        /* Prepend access modifiers to parameter declaration */
        ymir_1.Replacement.append(param.name.getStart(), member.modifiers ? member.modifiers.map((mod) => mod.getText()).join(' ') + ' ' : 'public '),
    ];
}
function getEndOfDirectivesAndSuper(construct) {
    const firstStmtIndex = Math.max(getFinalDirectiveIndex(construct), getSuperCallIndex(construct));
    if (firstStmtIndex === -1 || construct.body.statements.length === 0)
        return construct.body.getStart() + 1;
    return construct.body.statements[firstStmtIndex].end;
}
/**
 * If directive(s) such as 'use strict' exist in the constructor body, returns
 * the last directive, otherwise undefined.
 */
function getFinalDirectiveIndex(construct) {
    let finalDirectiveIndex = -1;
    for (const [index, stmt] of construct.body.statements.entries()) {
        if (tsutils_1.isExpressionStatement(stmt) && tsutils_1.isStringLiteral(stmt.expression)) {
            finalDirectiveIndex = index;
            continue;
        }
        break;
    }
    return finalDirectiveIndex;
}
/**
 * Takes a parameter property and returns the text of the equivalent class element.
 */
function getPropertyTextFromParam(param) {
    const modifiers = param.modifiers ? param.modifiers.map((mod) => mod.getText()).join(' ') : '';
    const typeAnnotations = param.type ? param.type.getText() : '';
    return (modifiers +
        ' ' +
        param.name.getText() +
        (param.questionToken ? '?' : '') +
        (typeAnnotations ? ': ' + typeAnnotations : '') +
        ';');
}
/**
 * Returns the super call expression index if it exists, otherwise -1.
 */
function getSuperCallIndex(construct) {
    const statementIndex = getFinalDirectiveIndex(construct) + 1;
    const statements = construct.body.statements;
    return statementIndex === statements.length || !isStatementSuperCall(statements[statementIndex])
        ? -1
        : statementIndex;
}
/**
 * Returns true when the prop is assigned any value other than the param identifier.
 */
function isNonParamAssignmentToProp(stmt, param) {
    return (tsutils_1.isBinaryExpression(stmt.expression) &&
        tsutils_1.isPropertyAccessExpression(stmt.expression.left) &&
        stmt.expression.left.expression.kind === ts.SyntaxKind.ThisKeyword &&
        tsutils_1.getPropertyName(stmt.expression.left.name) === param.name.getText() &&
        (!tsutils_1.isIdentifier(stmt.expression.right) ||
            (tsutils_1.isIdentifier(stmt.expression.right) && stmt.expression.right.text !== param.name.getText())));
}
function isPropertyMatchForParam(mem, param) {
    return (tsutils_1.isPropertyDeclaration(mem) &&
        !mem.initializer &&
        (!mem.questionToken ? !param.questionToken : !!param.questionToken) &&
        tsutils_1.getPropertyName(mem.name) === param.name.getText() &&
        (mem.type ? (param.type && param.type.getText()) === mem.type.getText() : !param.type));
}
/**
 * Checks if a statement assigns the parameter to a class member property.
 */
function isSimpleParamToPropAssignment(stmt, param) {
    return (tsutils_1.isExpressionStatement(stmt) &&
        tsutils_1.isBinaryExpression(stmt.expression) &&
        tsutils_1.isPropertyAccessExpression(stmt.expression.left) &&
        stmt.expression.left.expression.kind === ts.SyntaxKind.ThisKeyword &&
        tsutils_1.isIdentifier(stmt.expression.right) &&
        stmt.expression.left.name.text === param.name.getText() &&
        stmt.expression.right.text === param.name.getText());
}
function isStatementSuperCall(stmt) {
    return (tsutils_1.isExpressionStatement(stmt) &&
        tsutils_1.isCallExpression(stmt.expression) &&
        stmt.expression.expression.kind === ts.SyntaxKind.SuperKeyword);
}
/**
 * We can only autofix longhand props -> param props if certain conditions are met within the constructor body.
 * See https://github.com/fimbullinter/wotan/issues/167#issuecomment-378945862 for more details on these conditions.
 */
function isStatementWithPossibleSideEffects(stmt, param) {
    return (!tsutils_1.isExpressionStatement(stmt) ||
        !tsutils_1.isBinaryExpression(stmt.expression) ||
        stmt.expression.operatorToken.kind !== ts.SyntaxKind.EqualsToken ||
        tsutils_1.hasSideEffects(stmt.expression.left) ||
        tsutils_1.hasSideEffects(stmt.expression.right) ||
        isNonParamAssignmentToProp(stmt, param));
}
/**
 * Returns all statements in a constructor body with the exception of super calls
 * and directives such as 'use strict'.
 */
function statementsMinusDirectivesAndSuper(construct) {
    return construct.body.statements.slice(Math.max(getFinalDirectiveIndex(construct), getSuperCallIndex(construct)) + 1);
}
//# sourceMappingURL=parameter-properties.js.map