"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
const utils_1 = require("../utils");
let Rule = class Rule extends ymir_1.ConfigurableRule {
    parseOptions(options) {
        let destructuring = 'all';
        if (options != undefined && options.destructuring === 'any')
            destructuring = 'any';
        return {
            destructuring,
        };
    }
    apply() {
        const declarationInfo = new Map();
        const listInfo = new Map();
        for (const [name, variableInfo] of tsutils_1.collectVariableUsage(this.sourceFile)) {
            if (variableInfo.inGlobalScope || variableInfo.exported)
                continue;
            for (const d of variableInfo.declarations) {
                let declaration = d.parent;
                if (tsutils_1.isBindingElement(declaration))
                    declaration = tsutils_1.getDeclarationOfBindingElement(declaration);
                if (!tsutils_1.isVariableDeclaration(declaration))
                    continue;
                const parent = declaration.parent;
                if (parent.kind !== ts.SyntaxKind.VariableDeclarationList)
                    continue;
                const kind = tsutils_1.getVariableDeclarationKind(parent);
                if (kind === tsutils_1.VariableDeclarationKind.Const)
                    continue;
                const canBeConst = (declaration.initializer !== undefined || tsutils_1.isForInOrOfStatement(parent.parent)) &&
                    !variableInfo.uses.some(utils_1.isVariableReassignment) &&
                    (kind === tsutils_1.VariableDeclarationKind.Let ||
                        variableInfo.declarations.length === 1 && !usedInOuterScopeOrTdz(name, variableInfo.uses));
                let list = listInfo.get(parent);
                if (list === undefined) {
                    list = {
                        fixable: canBeConst,
                        declarations: [],
                    };
                    listInfo.set(parent, list);
                }
                let decl = declarationInfo.get(declaration);
                if (decl === undefined) {
                    decl = {
                        identifiers: [],
                        fixable: canBeConst,
                    };
                    list.declarations.push(decl);
                    declarationInfo.set(declaration, decl);
                }
                if (canBeConst) {
                    decl.identifiers.push(name);
                }
                else {
                    decl.fixable = false;
                    list.fixable = false;
                }
            }
        }
        for (const [list, listMeta] of listInfo) {
            let fix;
            if (listMeta.fixable) {
                fix = ymir_1.Replacement.replace(list.declarations.pos - 3, list.declarations.pos, 'const');
            }
            else if (list.parent.kind === ts.SyntaxKind.ForStatement) {
                continue;
            }
            const keyword = this.sourceFile.text.substr(list.declarations.pos - 3, 3);
            for (const declaration of listMeta.declarations) {
                if (declaration.identifiers.length === 0 || !declaration.fixable && this.options.destructuring === 'all')
                    continue;
                for (const name of declaration.identifiers)
                    this.addFindingAtNode(name, `Variable '${name.text}' is never reassigned. Prefer 'const' instead of '${keyword}'.`, fix);
            }
        }
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
function getScopeBoundary(node) {
    do
        node = node.parent;
    while (!tsutils_1.isScopeBoundary(node));
    return node;
}
function getFunctionScopeBoundary(node) {
    while (!tsutils_1.isFunctionScopeBoundary(node))
        node = node.parent;
    return node;
}
function usedInOuterScopeOrTdz(declaration, uses) {
    const deadZones = [{ pos: 0, end: declaration.pos }];
    declaration = declaration.parent;
    while (tsutils_1.isBindingElement(declaration)) {
        if (declaration.initializer !== undefined)
            deadZones.push({ pos: declaration.initializer.pos, end: declaration.initializer.end });
        declaration = declaration.parent.parent;
    }
    const varDeclaration = declaration;
    if (varDeclaration.initializer !== undefined) {
        deadZones.push({ pos: varDeclaration.initializer.pos, end: varDeclaration.initializer.end });
    }
    else {
        deadZones.push({
            pos: varDeclaration.parent.parent.expression.pos,
            end: varDeclaration.parent.parent.expression.end
        });
    }
    const declaredScope = getScopeBoundary(declaration.parent);
    const functionScope = getFunctionScopeBoundary(declaredScope);
    for (const use of uses) {
        const useScope = getScopeBoundary(use.location.parent);
        if (useScope.pos < declaredScope.pos || useScope.end > declaredScope.end)
            return true;
        if ((use.domain & (tsutils_1.UsageDomain.Value | tsutils_1.UsageDomain.TypeQuery)) === tsutils_1.UsageDomain.Value) {
            if (getFunctionScopeBoundary(useScope) !== functionScope)
                return true; // maybe used before declaration
            for (const deadZone of deadZones)
                if (use.location.pos >= deadZone.pos && use.location.pos < deadZone.end)
                    return true; // used in temporal dead zone
        }
    }
    return false;
}
//# sourceMappingURL=prefer-const.js.map