"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
const path = require("path");
const utils_1 = require("../utils");
let Rule = class Rule extends ymir_1.TypedRule {
    get usage() {
        const usage = tsutils_1.collectVariableUsage(this.sourceFile);
        Object.defineProperty(this, 'usage', {
            value: usage,
            writable: false,
        });
        return usage;
    }
    apply() {
        for (const node of this.context.getFlatAst())
            if (node.kind === ts.SyntaxKind.ForStatement)
                this.checkForStatement(node);
    }
    checkForStatement(node) {
        if (node.initializer === undefined || node.condition === undefined || node.incrementor === undefined)
            return;
        const indexVariable = extractIndexVariable(node.initializer, node.incrementor);
        if (indexVariable === undefined)
            return;
        const arrayVariable = extractArrayVariable(node.condition, indexVariable.text);
        if (arrayVariable === undefined)
            return;
        if (!isReadonlyArrayAccess(this.usage.get(indexVariable).uses, arrayVariable.getText(this.sourceFile), node, this.sourceFile))
            return;
        if (this.isIterationPossible(arrayVariable))
            this.addFinding(node.getStart(this.sourceFile), node.statement.pos, `Prefer a 'for-of' loop over a 'for' loop for this simple iteration.`);
    }
    isIterationPossible(node) {
        const type = this.checker.getTypeAtLocation(node);
        return this.isIterationProtocolAvailable()
            ? this.isIterable(this.checker.getApparentType(type), node)
            : tsutils_1.unionTypeParts(this.checker.getBaseConstraintOfType(type) || type).every(this.isArrayLike, this);
    }
    isIterationProtocolAvailable() {
        return this.sourceFile.languageVersion >= ts.ScriptTarget.ES2015 || this.context.compilerOptions.downlevelIteration === true;
    }
    isArrayLike(type) {
        if (tsutils_1.isTypeReference(type))
            type = type.target;
        if (type.getNumberIndexType() === undefined)
            return false;
        if (type.flags & ts.TypeFlags.StringLike)
            return this.sourceFile.languageVersion >= ts.ScriptTarget.ES5; // iterating string is only possible starting from ES5
        if (type.symbol !== undefined && /^(Concat|Readonly)?Array$/.test(type.symbol.escapedName) &&
            type.symbol.declarations !== undefined && type.symbol.declarations.some(this.isDeclaredInDefaultLib, this))
            return true;
        if (tsutils_1.isIntersectionType(type))
            return type.types.some(this.isArrayLike, this);
        const baseTypes = type.getBaseTypes();
        return baseTypes !== undefined && baseTypes.some(this.isArrayLike, this);
    }
    isDeclaredInDefaultLib(node) {
        // we assume it's the global array type if it comes from any lib.xxx.d.ts file
        return path.normalize(path.dirname(node.getSourceFile().fileName))
            === path.dirname(ts.getDefaultLibFilePath(this.context.compilerOptions));
    }
    isIterable(type, node) {
        const indexType = type.getNumberIndexType() || type.getStringIndexType();
        if (indexType === undefined)
            return false;
        const iteratorFn = type.getProperties().find((p) => p.escapedName === '__@iterator');
        if (!isPresentPublicAndRequired(iteratorFn))
            return false;
        return checkReturnTypeAndRequireZeroArity(this.checker.getTypeOfSymbolAtLocation(iteratorFn, node), (iterator) => {
            const next = iterator.getProperty('next');
            return isPresentPublicAndRequired(next) &&
                checkReturnTypeAndRequireZeroArity(this.checker.getTypeOfSymbolAtLocation(next, node), (iteratorResult) => {
                    const done = iteratorResult.getProperty('done');
                    if (!isPresentPublicAndRequired(done) ||
                        !tsutils_1.unionTypeParts(this.checker.getTypeOfSymbolAtLocation(done, node))
                            .every((t) => tsutils_1.isTypeFlagSet(t, ts.TypeFlags.BooleanLike)))
                        return false;
                    const value = iteratorResult.getProperty('value');
                    return isPresentPublicAndRequired(value) &&
                        utils_1.typesAreEqual(this.checker.getTypeOfSymbolAtLocation(value, node), indexType, this.checker);
                });
        });
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
function checkReturnTypeAndRequireZeroArity(type, cb) {
    let zeroArity = false;
    for (const signature of type.getCallSignatures()) {
        if (!cb(signature.getReturnType()))
            return false;
        if (signatureHasArityZero(signature))
            zeroArity = true;
    }
    return zeroArity;
}
function signatureHasArityZero(signature) {
    if (signature.parameters.length === 0)
        return true;
    const decl = signature.parameters[0].declarations[0];
    return decl !== undefined && isOptionalParameter(decl);
}
function isOptionalParameter(node) {
    if (node.questionToken !== undefined || node.dotDotDotToken !== undefined)
        return true;
    if (node.flags & ts.NodeFlags.JavaScriptFile && ts.getJSDocParameterTags(node).some((tag) => tag.isBracketed))
        return true;
    if (node.initializer === undefined)
        return false;
    const parameters = node.parent.parameters;
    const nextIndex = parameters.indexOf(node) + 1;
    if (nextIndex === parameters.length)
        return true;
    return isOptionalParameter(parameters[nextIndex]);
}
function isPresentPublicAndRequired(symbol) {
    return symbol !== undefined && !tsutils_1.isSymbolFlagSet(symbol, ts.SymbolFlags.Optional) &&
        (symbol.declarations === undefined ||
            symbol.declarations.every((d) => (ts.getCombinedModifierFlags(d) & ts.ModifierFlags.NonPublicAccessibilityModifier) === 0));
}
function isReadonlyArrayAccess(uses, arrayVariable, statement, sourceFile) {
    let arrayAccess = false;
    for (const { location: use } of uses) {
        if (use.pos < statement.pos || use.end > statement.end)
            return false; // used outside of the loop
        if (use.pos < statement.statement.pos)
            continue; // uses in loop header are already checked
        const parent = use.parent;
        if (!tsutils_1.isElementAccessExpression(parent) ||
            parent.argumentExpression !== use ||
            parent.expression.getText(sourceFile) !== arrayVariable ||
            tsutils_1.getAccessKind(parent) & tsutils_1.AccessKind.Modification)
            return false;
        arrayAccess = true;
    }
    return arrayAccess; // needs at least one array access
}
function extractArrayVariable(condition, indexVariable) {
    if (!tsutils_1.isBinaryExpression(condition))
        return;
    let left;
    let right;
    switch (condition.operatorToken.kind) {
        case ts.SyntaxKind.LessThanToken:
            // i < foo.length
            ({ left, right } = condition);
            break;
        case ts.SyntaxKind.GreaterThanToken:
            // foo.length > i
            ({ left: right, right: left } = condition);
            break;
        default:
            return;
    }
    if (!isVariable(left, indexVariable) || !tsutils_1.isPropertyAccessExpression(right) || right.name.text !== 'length')
        return;
    return right.expression;
}
function extractIndexVariable(initializer, incrementor) {
    // there must be only one variable declared
    if (!tsutils_1.isVariableDeclarationList(initializer) ||
        initializer.declarations.length !== 1)
        return;
    const declaration = initializer.declarations[0];
    // variable must be initialized to 0
    if (declaration.name.kind !== ts.SyntaxKind.Identifier || declaration.initializer === undefined ||
        !isNumber(declaration.initializer, 0))
        return;
    // variable must be incremented by one
    if (!isIncrementedByOne(incrementor, declaration.name.text))
        return;
    return declaration.name;
}
function isIncrementedByOne(node, name) {
    if (tsutils_1.isPostfixUnaryExpression(node) || tsutils_1.isPrefixUnaryExpression(node))
        // ++var or var++
        return node.operator === ts.SyntaxKind.PlusPlusToken && isVariable(node.operand, name);
    if (!tsutils_1.isBinaryExpression(node))
        return false;
    switch (node.operatorToken.kind) {
        case ts.SyntaxKind.PlusEqualsToken:
            // var += 1
            return isVariable(node.left, name) && isNumber(node.right, 1);
        case ts.SyntaxKind.EqualsToken:
            // var = var + 1 or var = 1 + var
            return isVariable(node.left, name) &&
                tsutils_1.isBinaryExpression(node.right) && node.right.operatorToken.kind === ts.SyntaxKind.PlusToken &&
                (isVariable(node.right.left, name) && isNumber(node.right.right, 1) ||
                    isNumber(node.right.left, 1) && isVariable(node.right.right, name));
        default:
            return false;
    }
}
function isVariable(node, name) {
    return tsutils_1.isIdentifier(node) && node.text === name;
}
function isNumber(node, value) {
    return tsutils_1.isNumericLiteral(node) && node.text === String(value);
}
//# sourceMappingURL=prefer-for-of.js.map