"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
const utils_1 = require("../utils");
let Rule = class Rule extends ymir_1.TypedRule {
    apply() {
        const re = /(?:[.\n]|\*\/)\s*assign\b/g;
        let wrappedAst;
        for (let match = re.exec(this.sourceFile.text); match !== null; match = re.exec(this.sourceFile.text)) {
            const { node } = tsutils_1.getWrappedNodeAtPosition(wrappedAst || (wrappedAst = this.context.getWrappedAst()), re.lastIndex - 1);
            if (node.kind !== ts.SyntaxKind.Identifier || node.end !== re.lastIndex)
                continue;
            const parent = node.parent;
            if (!tsutils_1.isPropertyAccessExpression(parent) || parent.name !== node ||
                !tsutils_1.isIdentifier(parent.expression) || parent.expression.text !== 'Object')
                continue;
            const grandParent = parent.parent;
            if (!tsutils_1.isCallExpression(grandParent) || grandParent.expression !== parent ||
                grandParent.arguments.length === 0 || !tsutils_1.isObjectLiteralExpression(grandParent.arguments[0]))
                continue;
            if (grandParent.arguments.length === 1) {
                this.addFindingAtNode(grandParent, "No need for 'Object.assign', use the object directly.", createFix(grandParent, this.sourceFile));
            }
            else if (grandParent.arguments.every(this.isSpreadableObject, this)) {
                this.addFindingAtNode(grandParent, "Prefer object spread over 'Object.assign'.", createFix(grandParent, this.sourceFile));
            }
        }
    }
    isSpreadableObject(node) {
        switch (node.kind) {
            case ts.SyntaxKind.ThisKeyword:
            case ts.SyntaxKind.SpreadElement:
                return false;
            case ts.SyntaxKind.ObjectLiteralExpression:
                return true;
        }
        const type = this.checker.getTypeAtLocation(node);
        if (type.flags & (ts.TypeFlags.Any | ts.TypeFlags.Unknown))
            return true;
        let seenObject = false;
        for (const t of tsutils_1.unionTypeParts(type)) {
            if (t.flags & (ts.TypeFlags.Object | ts.TypeFlags.NonPrimitive)) {
                seenObject = true;
            }
            else if (!tsutils_1.isFalsyType(t)) {
                return false;
            }
        }
        return seenObject;
    }
};
Rule = tslib_1.__decorate([
    ymir_1.excludeDeclarationFiles
], Rule);
exports.Rule = Rule;
function createFix(node, sourceFile) {
    const args = node.arguments;
    const objectNeedsParens = utils_1.objectLiteralNeedsParens(node);
    const fix = [
        ymir_1.Replacement.replace(node.getStart(sourceFile), args[0].getStart(sourceFile), `${objectNeedsParens ? '(' : ''}{`),
        ymir_1.Replacement.replace(node.end - 1, node.end, `}${objectNeedsParens ? ')' : ''}`),
    ];
    let removedPrevious = false;
    for (let i = 0; i < args.length; ++i) {
        const arg = args[i];
        if (!tsutils_1.isObjectLiteralExpression(arg)) {
            fix.push(ymir_1.Replacement.append(arg.getStart(sourceFile), '...'));
            removedPrevious = false;
            continue;
        }
        if (arg.properties.length === 0) {
            let end = arg.end;
            if (i !== args.length - 1) {
                end = args[i + 1].getStart(sourceFile);
            }
            else if (args.hasTrailingComma) {
                end = args.end;
            }
            // remove empty object iteral and the following comma if exists
            fix.push(ymir_1.Replacement.delete(removedPrevious ? arg.getStart(sourceFile) : arg.pos, end));
            removedPrevious = true;
        }
        else {
            const start = arg.getStart(sourceFile);
            fix.push(
            // remove open brace
            ymir_1.Replacement.delete(start, start + 1), 
            // remove trailing comma if exists and close brace
            ymir_1.Replacement.delete(arg.properties[arg.properties.length - 1].end, arg.end));
            removedPrevious = false;
        }
    }
    return fix;
}
//# sourceMappingURL=prefer-object-spread.js.map