"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ymir_1 = require("@fimbul/ymir");
const ts = require("typescript");
const tsutils_1 = require("tsutils");
const utils_1 = require("../utils");
let Rule = class Rule extends ymir_1.ConfigurableRule {
    parseOptions(options) {
        return {
            style: options && options.style === 'classic' ? 'classic' : 'as',
        };
    }
    apply() {
        return this.options.style === 'classic' ? enforceClassicTypeAssertion(this.context) : enforceAsTypeAssertion(this.context);
    }
};
Rule = tslib_1.__decorate([
    ymir_1.typescriptOnly,
    ymir_1.excludeDeclarationFiles,
    ymir_1.predicate((sourceFile) => sourceFile.languageVariant === ts.LanguageVariant.Standard || 'excludes JSX files')
], Rule);
exports.Rule = Rule;
function enforceClassicTypeAssertion(context) {
    const re = /\bas\b/g;
    let wrappedAst;
    for (let match = re.exec(context.sourceFile.text); match !== null; match = re.exec(context.sourceFile.text)) {
        const { node } = tsutils_1.getWrappedNodeAtPosition(wrappedAst || (wrappedAst = context.getWrappedAst()), match.index);
        if (!tsutils_1.isAsExpression(node) || node.type.pos !== re.lastIndex)
            continue;
        const parent = node.parent;
        const expressionNeedsParens = node.expression.kind === ts.SyntaxKind.BinaryExpression;
        const needsParens = tsutils_1.isBinaryExpression(parent) && parent.operatorToken.kind === ts.SyntaxKind.AsteriskAsteriskToken;
        context.addFinding(match.index, node.end, "Use the classic type assertion style '<T>obj' instead.", [
            ymir_1.Replacement.append(node.getStart(context.sourceFile), `${charIf(needsParens, '(')}<${node.type.getText(context.sourceFile)}>${charIf(expressionNeedsParens, '(')}`),
            ymir_1.Replacement.replace(node.expression.end, node.end, charIf(expressionNeedsParens, ')') + charIf(needsParens, ')')),
        ]);
    }
}
function enforceAsTypeAssertion(context) {
    for (const node of context.getFlatAst()) {
        if (tsutils_1.isTypeAssertion(node)) {
            const assertionParens = assertionNeedsParens(node);
            const expressionParens = node.expression.kind === ts.SyntaxKind.YieldExpression ||
                !assertionParens && utils_1.expressionNeedsParensWhenReplacingNode(node.expression, node);
            const start = node.getStart(context.sourceFile);
            context.addFinding(start, node.expression.pos, "Use 'obj as T' instead.", [
                ymir_1.Replacement.replace(start, node.expression.getStart(context.sourceFile), charIf(expressionParens, '(') + charIf(assertionParens, '(')),
                ymir_1.Replacement.append(node.end, `${charIf(expressionParens, ')')} as ${node.type.getText(context.sourceFile)}${charIf(assertionParens, ')')}`),
            ]);
        }
    }
}
function charIf(condition, char) {
    return condition ? char : '';
}
function assertionNeedsParens(node) {
    let parent = node.parent;
    // fixing unary expression like 'await <T>foo' to 'await foo as T' asserts the result of the entire expression, not just the operand
    switch (parent.kind) {
        case ts.SyntaxKind.PrefixUnaryExpression:
        case ts.SyntaxKind.TypeOfExpression:
        case ts.SyntaxKind.AwaitExpression:
        case ts.SyntaxKind.VoidExpression:
            return true;
    }
    // fixing '<T>foo & bar' to 'foo as T & bar' would parse 'T & bar' as intersection type, therefore we need to add parens
    while (tsutils_1.isBinaryExpression(parent)) {
        if (node === parent.left) {
            switch (parent.operatorToken.kind) {
                case ts.SyntaxKind.AmpersandToken:
                case ts.SyntaxKind.BarToken:
                    return true;
                default:
                    return false;
            }
        }
        node = parent;
        parent = node.parent;
    }
    return false;
}
//# sourceMappingURL=type-assertion.js.map