"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const tsutils_1 = require("tsutils");
function* switchStatements(context) {
    const { text } = context.sourceFile;
    const re = /\bswitch\s*[(/]/g;
    let wrappedAst;
    for (let match = re.exec(text); match !== null; match = re.exec(text)) {
        const { node } = tsutils_1.getWrappedNodeAtPosition(wrappedAst || (wrappedAst = context.getWrappedAst()), match.index);
        if (node.kind === ts.SyntaxKind.SwitchStatement && node.getStart(context.sourceFile) === match.index)
            yield node;
    }
}
exports.switchStatements = switchStatements;
function* tryStatements(context) {
    const { text } = context.sourceFile;
    const re = /\btry\s*[{/]/g;
    let wrappedAst;
    for (let match = re.exec(text); match !== null; match = re.exec(text)) {
        const { node } = tsutils_1.getWrappedNodeAtPosition(wrappedAst || (wrappedAst = context.getWrappedAst()), match.index);
        if (node.kind === ts.SyntaxKind.TryStatement && node.tryBlock.pos - 'try'.length === match.index)
            yield node;
    }
}
exports.tryStatements = tryStatements;
function isAsyncFunction(node) {
    switch (node.kind) {
        case ts.SyntaxKind.FunctionDeclaration:
        case ts.SyntaxKind.MethodDeclaration:
            if (node.body === undefined)
                return false;
        // falls through
        case ts.SyntaxKind.ArrowFunction:
            if (node.body.kind !== ts.SyntaxKind.Block)
                return false;
        // falls through
        case ts.SyntaxKind.FunctionExpression:
            break;
        default:
            return false;
    }
    return tsutils_1.hasModifier(node.modifiers, ts.SyntaxKind.AsyncKeyword);
}
exports.isAsyncFunction = isAsyncFunction;
function isVariableReassignment(use) {
    return (use.domain & (tsutils_1.UsageDomain.Value | tsutils_1.UsageDomain.TypeQuery)) === tsutils_1.UsageDomain.Value && tsutils_1.isReassignmentTarget(use.location);
}
exports.isVariableReassignment = isVariableReassignment;
function* childStatements(node) {
    switch (node.kind) {
        case ts.SyntaxKind.IfStatement:
            yield node.thenStatement;
            if (node.elseStatement !== undefined)
                yield node.elseStatement;
            break;
        case ts.SyntaxKind.ForStatement:
        case ts.SyntaxKind.ForOfStatement:
        case ts.SyntaxKind.ForInStatement:
        case ts.SyntaxKind.WhileStatement:
        case ts.SyntaxKind.DoStatement:
        case ts.SyntaxKind.LabeledStatement:
        case ts.SyntaxKind.WithStatement:
            yield node.statement;
            break;
        case ts.SyntaxKind.SwitchStatement:
            for (const clause of node.caseBlock.clauses)
                yield* clause.statements;
            break;
        case ts.SyntaxKind.Block:
            yield* node.statements;
            break;
        case ts.SyntaxKind.TryStatement:
            yield* node.tryBlock.statements;
            if (node.catchClause !== undefined)
                yield* node.catchClause.block.statements;
            if (node.finallyBlock !== undefined)
                yield* node.finallyBlock.statements;
    }
}
exports.childStatements = childStatements;
function getLeadingExpressionWithPossibleParsingAmbiguity(expr) {
    switch (expr.kind) {
        case ts.SyntaxKind.PropertyAccessExpression:
        case ts.SyntaxKind.ElementAccessExpression:
        case ts.SyntaxKind.AsExpression:
        case ts.SyntaxKind.CallExpression:
        case ts.SyntaxKind.NonNullExpression:
            return expr.expression;
        case ts.SyntaxKind.PostfixUnaryExpression:
            return expr.operand;
        case ts.SyntaxKind.BinaryExpression:
            return expr.left;
        case ts.SyntaxKind.ConditionalExpression:
            return expr.condition;
        case ts.SyntaxKind.TaggedTemplateExpression:
            return expr.tag;
        default:
            return;
    }
}
function expressionNeedsParensWhenReplacingNode(expr, replaced) {
    // this currently doesn't handle the following cases
    // (yield) as any
    // await (yield)
    // (1).toString()
    // ({foo} = {foo: 1});
    // binary operator precendence
    while (true) {
        switch (expr.kind) {
            case ts.SyntaxKind.ObjectLiteralExpression:
            case ts.SyntaxKind.FunctionExpression:
            case ts.SyntaxKind.ClassExpression:
                return parentRequiresParensForNode(expr.kind, replaced);
            default:
                const current = getLeadingExpressionWithPossibleParsingAmbiguity(expr);
                if (current === undefined)
                    return false;
                expr = current;
        }
    }
}
exports.expressionNeedsParensWhenReplacingNode = expressionNeedsParensWhenReplacingNode;
function parentRequiresParensForNode(kind, replaced) {
    while (true) {
        const parent = replaced.parent;
        switch (parent.kind) {
            case ts.SyntaxKind.ArrowFunction:
                return kind === ts.SyntaxKind.ObjectLiteralExpression;
            case ts.SyntaxKind.ExpressionStatement:
                return true;
            case ts.SyntaxKind.ExportAssignment:
                return !parent.isExportEquals && kind !== ts.SyntaxKind.ObjectLiteralExpression;
        }
        if (getLeadingExpressionWithPossibleParsingAmbiguity(parent) !== replaced)
            return false;
        replaced = parent;
    }
}
function objectLiteralNeedsParens(replaced) {
    return parentRequiresParensForNode(ts.SyntaxKind.ObjectLiteralExpression, replaced);
}
exports.objectLiteralNeedsParens = objectLiteralNeedsParens;
const typeFormat = ts.TypeFormatFlags.NoTruncation
    | ts.TypeFormatFlags.UseFullyQualifiedType
    | ts.TypeFormatFlags.WriteClassExpressionAsTypeLiteral
    | ts.TypeFormatFlags.UseStructuralFallback;
function typesAreEqual(a, b, checker) {
    return a === b || checker.typeToString(a, undefined, typeFormat) === checker.typeToString(b, undefined, typeFormat);
}
exports.typesAreEqual = typesAreEqual;
function* elementAccessSymbols(node, checker) {
    const { argumentExpression } = node;
    if (argumentExpression === undefined || argumentExpression.pos === argumentExpression.end)
        return;
    const { names } = tsutils_1.getLateBoundPropertyNames(argumentExpression, checker);
    if (names.length === 0)
        return;
    yield* propertiesOfType(checker.getApparentType(checker.getTypeAtLocation(node.expression)), names);
}
exports.elementAccessSymbols = elementAccessSymbols;
function* propertiesOfType(type, names) {
    for (const { symbolName, displayName } of names) {
        const symbol = tsutils_1.getPropertyOfType(type, symbolName);
        if (symbol !== undefined)
            yield { symbol, name: displayName };
    }
}
exports.propertiesOfType = propertiesOfType;
function hasDirectivePrologue(node) {
    switch (node.kind) {
        case ts.SyntaxKind.SourceFile:
        case ts.SyntaxKind.ModuleBlock:
            return true;
        case ts.SyntaxKind.Block:
            switch (node.parent.kind) {
                case ts.SyntaxKind.ArrowFunction:
                case ts.SyntaxKind.FunctionExpression:
                case ts.SyntaxKind.FunctionDeclaration:
                case ts.SyntaxKind.MethodDeclaration:
                case ts.SyntaxKind.Constructor:
                case ts.SyntaxKind.GetAccessor:
                case ts.SyntaxKind.SetAccessor:
                    return true;
                default:
                    return false;
            }
        default:
            return false;
    }
}
exports.hasDirectivePrologue = hasDirectivePrologue;
function formatPseudoBigInt(v) {
    return `${v.negative ? '-' : ''}${v.base10Value}n`;
}
exports.formatPseudoBigInt = formatPseudoBigInt;
//# sourceMappingURL=utils.js.map